from nuctools import *
import numpy as np
import pandas as pd
from matplotlib.pyplot import *
from numpy.testing import assert_array_equal,assert_allclose
import os


def test_sum_tof():
    directory = os.getcwd() + '/tests/test_files/'
    pa_answer = pd.read_csv(directory+'pa_la.csv').to_numpy()
    tof_answer = pd.read_csv(directory+'tof_la.csv')[['uwadc1','uwadc2']].to_numpy()

    filelist = [directory+'la_fp4a_t03_r03_0022.lst']
    tof_df,pa_df = sum_tof(directory+'la-agl.json',filelist)
    tof_uw_df = tof_df[['uwadc1','uwadc2']]
    
    assert_allclose(pa_answer,pa_df.to_numpy())
    assert_allclose(tof_answer,tof_uw_df.to_numpy())




import pandas as pd


__all__ = ["read_tab1_ascii"]

def read_tab1_ascii(filename,mat,mf,mt,minus1=False):
    """
    Read Tab1 file in ascii format from AMPX (translated to ascii from charmin)

    Parameters
    ----------
    filename : str 
        The name of the file to be read
    mat : int
        A four digit material MAT id defined by ENDF (e.g. 7328 for Ta-181)
    mf : int
        A two digit file MF id defined by ENDF
    mt : int
        A three digit reaction MT id defined by ENDF
    minus1 : bool
        Whether the length record is wrong and needs to subtract one

    Returns
    -------
    tab1 : DataFrame
        A pandas DataFrame of energy and cross section (1-D)

    Examples
    --------
    >>> import nuctools as nuc
    >>> mat,mf,mt = 7328,3,1
    >>> filename = "mytab1.dat"
    >>> tab1 = nuc.read_tab1_ascii(filename,mat,mf,mt)

    """
    string_to_match = "{:4d}{:2d}{:3d}    5".format(mat,mf,mt)

    skip,end = 0,0
    with open(filename,'r+') as f:
        filelist = f.readlines()
    for i,line in enumerate(filelist):
        if string_to_match in line:
            skip = i+1
            end = int(line.split()[0])
            break
    if end!=0:
        if minus1:
            end -= 1
        # now on line after match
        try:
            tab1 = pd.read_csv(filename,skiprows=skip,nrows=end,names=['e','cs'],delim_whitespace=True)
        except pd.errors.ParserError as err:
            print("string to match: \"{}\"".format(string_to_match))
            print("lines to skip: ",skip)
            print("lines to read: ",end)
            print(f"Unexpected {err=}, {type(err)=}")
            raise
    return tab1


import numpy as np
import matplotlib.pyplot as plt
import h5py as h5
from . import tof_tools
import os,sys

__all__ = ["sample",'get_montrig','count_rate','dtc','inc']

class sample:
    """
    Store all relevant information for a sample being 
    studied at the RPI Linac. 

    Stores and calculates data about a sample studied at
    the RPI Linac and follows the file structure and naming 
    conventions that are found in codes originating from 
    Brian McDermott, such as "40mprocess" and "h5query".
    The output from these codes is used as input for this
    class.

    Parameters
    ----------
    samp_name : str
        Name of sample of interest, e.g. Ta-2mm, MUST be 
        the same as it appears in file structure of initial
        raw binary files as it is typically propagated 
        through to the HDF5 file.
    data_type : str
        Type of data which was collected, e.g. yield,
        or transmission
    tof_arr : numpy array
        Time of flight data, 1st column is ascending
        time bins [us], 2nd column is bin width, 3rd
        column is the summed counts
    hdf_address : str, optional
        File address of the hdf file output from 40mprocess
        which contains: monitors, triggers, and pulse events;
        must end with slash
    phys_dims : array_like, optional
        Physical dimensions [cm] of sample being studied, must be 
        in format [thickness,length,width] or [thickness,radius]
        where thickness is the length of sample parallel to the
        Linac beam
    tau : float, optional
        The dead time of the system [us] used to collect the counts
        data, default value is 0 
    paralyzeable : bool, optional
        If system used to collect counts data was paralyzeable 
        a different dead time model must be used, default is 
        non-paralyzeable dead-time model
    plot_montrig : bool, optional
        Plot the monitor ratio through the montrig function
    factor : 1-d numpy array, optional
        Compression factors for the tof groups, 1st compression factor
        corresponds to time of flights below c_pt[0], 2nd corresponds to tofs above
        c_pt[0] and below c_pt[1], and so on until end of range is found.
    c_pts : 1-d numpy array, optional
        Compression time of flights that seperate the bins that are to be grouped.
        List points from lowest to highest.
    avg : bool, optional
        Return an average of group values instead of
        the sum in the group. group error is 1/N*(sum(error^2))^(1/2)
    binedge : bool, optional
        Return TOF column for grouped data as the first
        tof in group instead of the mean tof of the
        group.
    verbose : bool, optional
        Choose whether to print out approximate comp pt energy locations
        and indices. Default is False.
    FP : float, optional
        Flight path needed to convert the time of flight values of bins to
        energies for the compression point selection.
    verbose : bool, optional
        Option to print out extra information

    Notes
    -----
    Will eventually require break points and grouping 
    metadata.

    Allows for yield calculations, must later include 
    transmission.
    """
    def __init__(self,samp_name,data_type,tof_arr,hdf_address=None,
                 phys_dims=None,tau=0.0,paralyzeable=False,
                 plot_montrig=False,factor=None,c_pts=None,
                 avg=False,binedge=False,FP=None,verbose=False):
        
        if verbose == True:
            print("--------\n {} data for {}\n--------".format(data_type,
                  samp_name))

        self.samp_name = samp_name
        self.phys_dims = phys_dims
        self.data_type = data_type

        #------------------------------------------------------------
        #------------ Monitors and triggers -------------------------
        #------------------------------------------------------------
        default_montrig = False

        if hdf_address == None:

            default_montrig = True
            print("----------\n----------")
            print("Monitors and triggers not given, cps calculated with")
            print("default 1e8 monitor and trigger counts.")
            print("----------\n----------")

            self.montrig = np.array([1e8,1e8])

        else:
            self.montrig = get_montrig(hdf_address,samp_name,plot=plot_montrig,
                                       verbose=verbose)

        #------------------------------------------------------------
        #------------ Count rate ------------------------------------
        #------------------------------------------------------------
        if default_montrig:
            self.cps = count_rate(tof_arr,self.montrig[1],tau,paralyzeable=paralyzeable,
                 factor=factor,c_pts=c_pts,avg=avg,binedge=binedge,verbose=verbose,FP=FP)
        else: 
            self.cps = count_rate(tof_arr,self.montrig[1][1],tau,paralyzeable=paralyzeable,
                 factor=factor,c_pts=c_pts,avg=avg,binedge=binedge,verbose=verbose,FP=FP)
        #------------------------------------------------------------


#------------------------------------------------------------------------------

def get_montrig(hdf_address,samp_name,plot=False,verbose=False):
    """
    Collect monitors and triggers used to reduce
    capture data to proper dead time corrected
    monitor normalized count rates

    Parameters
    ----------
    hdf_address : str
        File address of the HDF5 file, must end with "/"
    samp_name : str
        Name of the sample being analyzed. Should
        match the sample name in the HDF5 filename
    plot : bool, optional
        Choose whether to plot the monitors and triggers
        as a function of cycle number
    verbose : bool, optional
        Choose whether to print out trigger and monitor 
        information.

    Returns
    -------
    montrig : numpy array
        Vector of monitor counts and triggers

    Notes
    -----
    The montrig array is printed out when function
    is called, user can see order of monitors and
    triggers.
    """

    with h5.File(hdf_address+samp_name+"_master.h5","r") as hdf:

        mon_counts = []
        det_counts = []
        sum_mon = np.zeros(8)
        sum_trig = np.zeros(2)

        for cycle in hdf[samp_name]:
            mon_vector = hdf[samp_name+"/"+cycle+"/MON"]
            trig_vector = hdf[samp_name+"/"+cycle+"/TRIG"][0]
            datag_array = hdf[samp_name+"/"+cycle+"/DATA/GOOD"]
            datat_array = hdf[samp_name+"/"+cycle+"/DATA/TRUNC"]
            sum_detcts = len(datag_array)+len(datat_array)
            mon_counts.append(mon_vector)
            det_counts.append(sum_detcts)

            # number of monitors stored in /MON != to 8
            if (len(sum_mon) != len(mon_vector)) or (len(sum_trig) != len(trig_vector)):
                raise ValueError("Unexpected number of mon's in /MON")

            sum_mon += mon_vector
            sum_trig += trig_vector

        num_cyc = len(hdf[samp_name])
        mon_counts = np.transpose(mon_counts)
        # monitor ratios 
        mr0,mr1,mr2 = np.zeros(num_cyc),np.zeros(num_cyc),np.zeros(num_cyc)
        for i in range(num_cyc):
            mr0[i] = det_counts[i]/mon_counts[0][i]
            mr1[i] = det_counts[i]/mon_counts[1][i]
            mr2[i] = det_counts[i]/mon_counts[2][i]
        # std of monitor cycles / mean of mon. cycles
        stdmean = [0,0,0]
        stdmean[0] = np.std(mr0)/np.mean(mr0)
        stdmean[1] = np.std(mr1)/np.mean(mr1)
        stdmean[2] = np.std(mr2)/np.mean(mr2)
        
        if plot:
            x = range(num_cyc)
            plt.plot(x,mr0/np.mean(mr0),label="Sum/Mon 0")
            plt.plot(x,mr1/np.mean(mr1),label="Sum/Mon 1")
            plt.plot(x,mr2/np.mean(mr2),label="Sum/Mon 2")
            plt.legend()
            plt.show()
        
        montrig = np.concatenate((sum_mon,sum_trig))
        if verbose == True:
            print("         Counts          Std/mean")
            for i in range(len(stdmean)):
                print("Mon. {} : {}  ,  {}".format(i,sum_mon[i],stdmean[i]))
            if len(sum_trig>1):
                print("Triggers : {}  {}".format(montrig[6],montrig[7]))

    return montrig


def count_rate(tof_arr,trig,tau=None,paralyzeable=False,factor=None,c_pts=None,
               avg=False,binedge=False,verbose=False,FP=None):
    """
    Converts counts to dead-time corrected count
    rates

    Parameters
    ----------
    tof_arr : 2-d numpy array
        Time of flight data, 1st column is ascending
        time bins [us], 2nd column is the summed counts,
        and the 3rd column (if desired) is error on the 
        counts. 3rd column only necessary if avg==True.
        !! time of flight data **MUST** be equal bin sizes !!
    trig : float
        Triggers recorded for the sample
    tau : float
        The dead time of the system used to collect 
        counts data [us]
    paralyzeable : bool,optional
        If the system used to collect the counts data
        was paralyzeable, a different dead time model
        must be used
    factor : 1-d numpy array, optional
        Compression factors for the tof groups, 1st compression factor
        corresponds to time of flights below c_pt[0], 2nd corresponds to tofs above
        c_pt[0] and below c_pt[1], and so on until end of range is found.
    c_pts : 1-d numpy array
        Compression time of flights that seperate the bins that are to be grouped.
        List points from lowest to highest.
    avg : bool, optional
        Return an average of group values instead of
        the sum in the group. group error is 1/N*(sum(error^2))^(1/2)
    binedge : bool, optional
        Return TOF column for grouped data as the first
        tof in group instead of the mean tof of the
        group.
    verbose : bool, optional
        Choose whether to print out approximate comp pt energy locations
        and indices. Default is False.
    FP : float
        Flight path needed to convert the time of flight values of bins to
        energies for the compression point selection.

    Returns
    -------
    dc_arr : 2-d numpy array
        TOF data, 1st column ascending time bins [us],
        2nd column is count rate [cps], 3rd column is 
        the error on that count rate

    Notes
    -----
    Units matter!!

    Error for count rates are sqrt of the counts vector user 
    provides.

    Dead time correction error given in the dead time correction
    function `dtc()`.

    .. math:: CPS = \\frac{C}{b_w Trig}

    .. math:: \\Delta C = \\sqrt{C}

    .. math:: \\Delta CPS = \\sqrt{\\left(\\frac{\\Delta C}{b_w Trig}\\right)^2}

    .. math:: CPS_{DC} = (CPS) dtcf 

    .. math:: \\Delta CPS_{DC}=\\sqrt{\\left(\\frac{\\Delta CPS}{CPS} \\right)^2 + \\left(\\frac{\\Delta dtcf}{dtcf} \\right)^2}

    """
    use_single_group = False
    use_comp_group = False

    # -------------------------------------------------------------------------
    # ---- run some checks ----------------------------------------------------
    # -------------------------------------------------------------------------
    # did the user give one grouping input but not the other?
    mismatch1 = ((factor is None) and (c_pts is not None))
    mismatch2 = ((factor is not None) and (c_pts is None))
    if mismatch1:
        raise ValueError("c_pts given, but factor is not.")
    # if mismatch2, we may need to use single group if factor is not an array
    if mismatch2:
        try:
            len(factor)
            raise ValueError()
        except TypeError:
            print('Using single group')
            use_single_group = True
        except:
            print('Factor is an array and c_pts is not given. Aborting.')
            raise ValueError("Check input for argument: factor.")
    if factor is not None and c_pts is not None:
        use_comp_group = True
    if use_comp_group:
        # we need the flight path for comp_group
        if FP is None:
            raise ValueError("Specify the flight path for comp_group()")

    # cast array into np array
    tof_arr = np.array(tof_arr)
    # if input is pandas, may need to transpose
    if tof_arr.shape[0] > tof_arr.shape[1]:
        tof_arr = np.transpose(tof_arr)

    # -------------------------------------------------------------------------
    # ---- Calculate and apply dead time correction ---------------------------
    # -------------------------------------------------------------------------
    if tau is None:
        tau = 0.0     # will be "corrected" by 1.0, with error 0.0

    tau = tau*1e-6    # [us --> s]

    if paralyzeable == False:
        bin_width = tof_arr[0][1]-tof_arr[0][0]
        dtcf = dtc(tof_arr[1],tau,trig,bin_width)
        dc_counts = tof_arr[1]*dtcf[0]
        # sum of the squares of relative errors in counts and dead-time corr.
        # time the corrected counts
        if avg:
            ecounts = tof_arr[2]
        else:
            ecounts = np.sqrt(tof_arr[1])

        rel_err_dtcf = (dtcf[1]/dtcf[0])
        rel_err_counts = (ecounts/tof_arr[1])
        edc_counts = np.sqrt(rel_err_dtcf**2 + rel_err_counts**2) * dc_counts
        # --------------------------------------------
        # check if error on correction is significant
        # --------------------------------------------
        ratio = rel_err_dtcf/(rel_err_dtcf+rel_err_counts)
        if (ratio > 0.02).any():
            print("Max % of dead time error: ",max(ratio))
            print("!! Dead time correction factor accounts for > 2% error in the counts !!")
            print("!! This is not accounted for in grouping if avg is not used !!")
            print("To fix, you have to change the hard-coded grouping function.")

    else:
        raise NotImplementedError("paralyzeable dead time not implemented.")

    # -------------------------------------------------------------------------
    # ---- Group the counts if user has specified -----------------------------
    # -------------------------------------------------------------------------
    if use_single_group:
        # include the input error on the counts
        gtof_array = tof_tools.single_group([tof_arr[0],dc_counts,ecounts],
            factor,avg=avg,binedge=binedge)
    if use_comp_group:
        # include the input error on the counts
        gtof_array = tof_tools.comp_group([tof_arr[0],dc_counts,ecounts],
            factor,c_pts,FP,avg=avg,binedge=binedge,verbose=verbose)

    # -------------------------------------------------------------------------
    # ---- Calculate the count rate -------------------------------------------
    # -------------------------------------------------------------------------
    if use_comp_group is False and use_single_group is False:
        cps = dc_counts/bin_width/trig/1e-6
        # error in the cps 
        err_cps = 1/bin_width/trig/1e-6 * edc_counts
        # stage a tof array for the return statement
        time_of_flight = tof_arr[0]
    else:
        cps = gtof_array[2]/gtof_array[1]/trig/1e-6
        # error in cps
        err_cps = 1/gtof_array[1]/trig/1e-6 * gtof_array[3]
        # stage a tof array for the return statement
        time_of_flight = gtof_array[0]

    cps_arr = np.array([time_of_flight,cps,err_cps])


    return cps_arr



def dtc(counts,dead_time,trigs,bin_width):
    """
    Calculate the TOF dead-time correction factor according
    to the non-paralyzable model

    Parameters
    ----------
    counts : array_like
        A 1-d array of counts as a function of tof (tof 
        in ascending order, equal bin spacing throughout)
    dead_time : float
        The dead time [us] associated with the data collection 
        system.
    trigs : int
        The number of triggers = the number of times channel 
        was collecting data. For RPI the number of LINAC 
        pulses.
    bin_width : float
        The width of the bins in time [us]. Should be constant
        for the array of counts.

    Returns
    -------
    dtcf : 2-d numpy array
        The dead time correction factor for the counts array
        based on the non-paralyzeable dead time model is in 
        column 0 of array dtcf, the error associated with it
        is in column 1. 
        len(dtcf[0]) = len(dtcf[1]) = len(counts).

    Notes
    -----
    dtc() returns 0's for the first bins that span the dead 
    time. 1's would keep the signal closer to reality, but 
    0's make it obvious to the user that those channels do 
    not have a correction applied, and data in this time of
    flight range is rarely used since it usually occurs before
    the gamma flash.

    .. math:: dtcf_{\\tau>b_w} = \\frac{1}{1-\\frac{SUM}{trig}} = \\frac{1}{1-\\frac{\\sum_iw_iC_i}{trig}}

    where the weights are all 1 except the first and last bin. 
    First weight is 0.5, last weight is the fraction of bins
    the dead time minus 0.5 covers (e.g. 3.1 minus 1/2 = 2.6) 
    minus the whole bins dead time minus 0.5 covers (2), so 
    the last weight is 0.6.

    Now the error.

    .. math:: \\Delta SUM = \\sqrt{\\sum_i(w_i\\sqrt{C_i})^2}

    .. math:: \\Delta dtcf_{\\tau>b_w} = \\sqrt{\\left(\\frac{\\partial dtcf_{\\tau>b_w}}{\\partial SUM}\\Delta SUM\\right)^2} = \\sqrt{\\left(\\frac{\\Delta SUM}{trig\\left(1-\\frac{SUM}{trig}\\right)^2}\\right)^2} 

    .. math:: = \\frac{\\Delta SUM}{trig\\left(1-\\frac{SUM}{trig}\\right)^2} = \\frac{\\Delta SUM}{trig}dtcf_{\\tau>b_w}^2

    If the dead time is less than the bin width:

    .. math:: dtcf_{\\tau<b_w} = \\frac{1}{1-\\frac{C\\tau}{b_wtrig}}

    .. math:: \\Delta dtcf_{\\tau<b_w} = \\sqrt{\\left(\\frac{\\partial dtcf_{\\tau<b_w}}{\\partial C}\\Delta C\\right)^2} = \\frac{\\tau\\sqrt{C}}{b_wtrig\\left(1-\\frac{C\\tau}{b_wtrig}\\right)^2}
    
    Algorithm is a Pythonized version of the dead-time-correction algorithm by Y. Danon, "Design and Construction of the RPI Enhanced Thermal Neutron Target and Thermal Cross Section Measurements of Rare Earth Isotopes.", Doctoral Thesis, (1993).
    """

    # cast as numpy = major speedup
    counts = np.array(counts)
    lc = len(counts)
    dtcf = np.zeros(lc)
    ddtcf = np.zeros(lc)
    if bin_width>=dead_time:
        x = counts*dead_time/bin_width/trigs
        dtcf = 1/(1-x)
        ddtcf = dead_time*np.sqrt(counts)/bin_width/trigs/(1-x)**2
    else:
        # how many bins does the dead time span - 0.5 a bin
        an = dead_time/bin_width-0.5
        # how many whole bins does the dead time span
        n = int(an)
        # bins 0-n will be left 0 
        print("Channels 0 -",n,"set to 0")
        # fraction of earliest bin that dead time spans
        f1 = an-n
        # add half of current bin
        SUM = counts[n+1:lc] * 0.5
        # (w_i(C_i)^0.5)^2 = C_i*0.5*0.5
        sSUM = SUM*0.5
        # add preceding whole bins within span of dead time
        for j in range(1,n+1):
            SUM += counts[n+1-j:lc-j]
            sSUM += counts[n+1-j:lc-j]
        # add the fraction of the last bin dead time spans
        SUM += f1*counts[0:lc-n-1]
        sSUM += f1**2*counts[0:lc-n-1]
        dSUM = np.sqrt(sSUM)

        dtcf[n+1:lc] = 1/(1-SUM/trigs)
        ddtcf[n+1:lc] = dSUM/trigs*dtcf[n+1:lc]**2
    return np.array([dtcf,ddtcf])


def inc(x):
    return x + 1








from ._version import *
from .agl_tools import *
from .ags_tools import *
from .ampx_tools import *
from .chem_tools import *
from .endf_tools import *
from .funky import *
from .hdf5_tools import *
from .math_tools import *
from .mcnp_tools import *
from .physics_tools import *
from .plot_tools import *
from .rpixdr_tools import *
from .sam_tools import *
from .sample_create import *
from .scaledata_tools import *
from .tof_tools import *
from .trans_tools import *
from .triton_tools import *
from .urr_tools import *



import numpy as np
import pandas as pd

__all__ = ["TAB1",'read_pendf_xs','write_pendf_xs',"write_endf_float","read_endf_float",
           "update_file2","setupdict","update_pardict","strip_line_num",
           "loglog_interp"]

class TAB1:
    """
    A class to mimic an ENDF-6 format TAB1 record

    Parameters
    ----------
    filename : str
        The file path to the ENDF file
    mat : int
        The ENDF format MAT number
    mf : int
        The ENDF format MF number
    mt : int
        The ENDF format MT number
    starting_line_no : int, optional
        The starting line number of the MAT, MF, and MT

    Attributes
    ----------
    interp_table : list
        The interpolation table; a list of 2-element lists with values of number 
        of points and interpolation scheme for those points. The sum of number
        of points should be equal to attribute `np`
    c1 : float
        Control record number 1
    c2 : float
        Control record number 2
    l1 : int
        integer record number 1
    l2 : int
        integer record number 1
    nr : int
        number of interpolation ranges
    np : int
        number of tabulated pairs
    data : DataFrame
        A DataFrame with attributes `e` and `cs` for pointwise cross section

    Methods
    -------
    read_tab1_cr
        Read the control record and interpolation tabel for a TAB1

    """
    def __init__(self,filename,mat,mf,mt,starting_line_no=0):
        self.mat = mat
        self.mf = mf
        self.mt = mt
        self.filename = filename
        self.starting_line_no = starting_line_no
        self.c1 = None
        self.c2 = None
        self.l1 = None
        self.l2 = None
        self.nr = None
        self.np = None
        self.interp_table = []

        self.read_tab1_cr()

        self.data = pd.DataFrame(read_pendf_xs(self.filename,self.starting_line_no,self.ending_line_no).T,
                                  columns=['e','cs'])

    def read_tab1_cr(self):
        """ designed for file 3 """
        start,numlines_nr,numlines_np = 0,0,0
        read_interp_table = False
        with open(self.filename,'r') as f:
            lines = f.readlines()
        for i,line in enumerate(lines):
            if i<self.starting_line_no:
                continue
            # find starting line
            if "{:4d}{:2d}{:3d}".format(self.mat,self.mf,self.mt) in line and start==0:
                start = i+1
                continue
            # read control record
            if i==start and i!=0:
                self.c1 = read_endf_float(line[0:11])
                self.c2 = read_endf_float(line[11:22])
                self.l1 = read_endf_float(line[22:33])
                self.l2 = read_endf_float(line[33:44])
                self.nr = read_endf_float(line[44:55])
                self.np = read_endf_float(line[55:66])
                numlines_nr = int(self.nr//3)
                if self.nr%3 > 0: 
                    numlines_nr +=1
                numlines_np = int(self.np//3)
                if self.np%3 > 0:
                    numlines_np +=1
                read_interp_table = True
            if read_interp_table == True and i<=start+numlines_nr and i > start:
                for j in range(3):
                    npi = read_endf_float(line[j*22:j*22+11])
                    inter = read_endf_float(line[j*22+11:j*22+22])
                    if npi!=0.0 and inter!=0.0:
                        self.interp_table.append([npi,inter])
            if self.nr is not None:
                if i> numlines_nr + start:
                    break
        self.starting_line_no = start+3
        self.ending_line_no = start+2+numlines_np


def read_pendf_xs(file,start,finish):
    """
    Read part of a PENDF format file.
    
    
    PENDF format files have sections of the file where the cross
    section is listed in 10 character exponential format as:
    
    energy1 cross_section1 energy2 cross_section2
    
    format. This function opens the file, grabs the specified lines
    and converts to floats and returns an energy array and a cross 
    section array.
    
    Parameters
    ----------
    file : str
        The full pathname for the file to be read, should be in PENDF
        format.
    start : int
        The line number to begin reading data
    finish : int 
        The line number to stop reading data

    Returns
    -------
	pointwise_cs : numpy array
	    An array that contains pointwise cross section: energies at
	    pointwise_cs[0], and the cross section values at 
	    pointwise_cs[1]
	
	Examples
	--------
	>>> import nuctools as nuc
	>>> file = '/Users/jesse/Desktop/temp.pendf'
	>>> start = 98
	>>> finish = 18984
	>>> energy,cross_section = nuc.read_3col_pendf(file,start,finish)

    
    """
    with open(file) as f:
        e = []
        cs = []

        break_outer = False

        for i,line in enumerate(f):
            # -------------------------------
            # Stop the loop once finish is reached
            # -------------------------------
            if i == finish:
                break
            if i >= start-1:
            	# -------------------------------
            	# Only include first 66 columns, split on space
            	# and convert to an array of strings
            	# -------------------------------
                word_len = 11
                word_start = 0
                for j in range(6):
                    word = line[word_start:word_start+11]

                    if( j%2 == 0 ):
                        # -------------------------------
                        # Grab the energies, convert to readable format
                        # -------------------------------
                        if( word == '           ' ):
                            break_outer = True
                            break # end of TAB1
                        e.append(read_endf_float(word))
                    else:
                        # -------------------------------
                        # Grab cross section, convert to readable format
                        # -------------------------------
                        cs.append(read_endf_float(word))
                    word_start+=word_len

            if( break_outer ):
                break # end of TAB1
    
    # -------------------------------
    # Convert to floats
    # -------------------------------
    e = np.array(e).astype(float)
    cs = np.array(cs).astype(float)

    # -------------------------------
    # Stack them into a numpy array
    # -------------------------------
    pointwise_cs = np.array([e,cs])
    
    return pointwise_cs


def write_pendf_xs(filename,energy,cs,mat_num,file_num,reaction_num):
    """
    Write energy and cross section into the PENDF format to be used 
    mostly for file 3 in the ENDF format.

    Parameters
    ----------
    filename : str
        The full file name of the file which is to be written
    energy : array-like
        The energies for each of the cross section points
    cs : array-like
        The cross section points to be written into the PENDF
        format
    mat_num : int
        The integer number (e.g. 7328 for Ta-181) that describes
        the material for which the data is written. This is 
        specified in the ENDF format 
    file_num : int
        The file number (most often to be 3) as specified by the
        ENDF format
    reaction_num : int
        The reaction number as specified by the ENDF format. For 
        example the total cross section reaction number is 1

    Returns
    -------
    no return : ascii file
        The text in a file in PENDF format that can be placed 
        into the full ENDF file of interest.

    """
    # ----------------------
    # Open new file
    # ----------------------
    with open(filename,'w+') as f:
        # ----------------------
        # Check if cross section or energy are NaN
        # ----------------------
        if (np.isnan(energy).any()) or (np.isnan(cs).any()):
            raise ValueError("Input energy or cross section contains NaN's")
        # ----------------------
        # Check if cross section is negative
        # ----------------------
        cs = np.array(cs)
        if cs[cs<0].any():
            raise ValueError("Input cross section is negative.")
        
        for i in range(len(energy)):
            # ----------------------
            # find exponent on power of 10
            # ----------------------
            exponent_e = int(np.floor(np.log10(energy[i])))
            if exponent_e < 0:
                sign_e = "-"
            else:
                sign_e = "+"

            if( cs[i] == 0.0 ):
                exponent_cs = 0
            else:
                exponent_cs = int(np.floor(np.log10(cs[i])))
            if exponent_cs < 0:
                sign_cs = "-"
            else:
                sign_cs = "+"
            # ----------------------
            # Write the energy and cross section points
            # ----------------------
            f.write(" {:0=7.6f}{}{} {:0=7.6f}{}{}".format(energy[i]/10**exponent_e,sign_e,abs(exponent_e),
                                                        cs[i]/10**exponent_cs,sign_cs,abs(exponent_cs)))
            # ----------------------
            # Tag the last line with ENDF info
            # ----------------------
            if (i+1)%3 == 0:
                f.write("{:d} {:d}  {:d}\n".format(mat_num,file_num,reaction_num))
            # ----------------------
            # Fill in space at the end
            # ----------------------
            if (i == len(energy)-1) and ((i+1)%3 != 0):
                how_many_gaps = 3-(i+1)%3
                spaces_per_gap = 22
                extra_space = " "*spaces_per_gap*how_many_gaps
                f.write(extra_space+"{:d} {:d}  {:d}\n".format(mat_num,file_num,reaction_num))


def write_endf_float(value):
    """
    Return the ENDF format string of a floating point number

    Parameters
    ----------
    value : float
        A single floating point value

    Returns
    -------
    valstring : str
        An ENDF format string of input par `value`
    """
    if(abs(value) < 1e-9 or abs(value) > 9.999e9):
        raise ValueError("value is too small or too big")
    valstring = "{:>13.6e}".format(value).replace('e','').replace('+0','+').replace('-0','-')
    # with AMPX written files we use "-0" instead of "+0" for some reason
    if( '+0' in valstring ):
        valstring = valstring.replace('+0','-0')
    return valstring

def read_endf_float(string):
    """
    Read the ENDF format string of a floating point number and return float

    Parameters
    ----------
    string : str
        A string for a single floating point value

    Returns
    -------
    value : float
        The value for the ENDF string float
    """
    if string.strip() == "":
        return 0.0
    elif "e" in string.lower():
        return float(string)
    elif "." in string:
        strsplit = string.split('.')
        return float(strsplit[0]+"."+strsplit[1].replace("-","e-").replace("+","e+"))
    else:
        return float(string.replace("-","e-").replace("+","e+"))

def update_file2(infile,outfile,energy_dict,mat):
    """
    Replace the res's in the infile with new values
    
    note: resonance must have unique energy

    Parameters
    ----------
    infile : str
        A file that includes ENDF-format File 2. This is the base 
        file that the outfile will match except updated values
    outfile : str
        File written from `infile` and updated resonance values
        from `energy_dict`
    energy_dict : dict
        A dictionary that uses keys based on the energies that can
        be identified **in the base file** `infile`
    mat : int
        The unique ENDF format MAT id

    Returns
    -------
    None : None
        Write a new file to `outfile`
    """

    with open(infile,'r+') as f:
        with open(outfile,'w+') as of:
            matfile = "{} {}".format(mat,2151)
            for line in f:
                #if(we're in file 2: res's)
                if(matfile in line):
                    for ekey in energy_dict:
                        #if( we find the res )
                        if ekey in line:
                            linelist = list(line)
                            for i,channel_pair in enumerate(energy_dict[ekey]):
                                channel_number = channel_pair[0]
                                channel_value = channel_pair[1]
                                start = 11*channel_number
                                end = 11+start
                                linelist[start:end] = endf_float_str(channel_value)
                            line = "".join(linelist)
                of.write(line)

def setupdict(parfile):
    """
    Set up a dictionary of varied parameters

    Only varied parameters are included in the dict
    
    Parameters
    ----------
    parfile : str
        The path to a par file that was used to run a SAMMY-like
        operation (normal, MC, etc.).

    Returns
    -------
    pardict : dict
        A dictionary with keys of res energies. Varied pars are
        added to a list for the appropriate res energy

    Notes
    -----
    ** Must match exact E-values of ENDF file that dict will be used to update **

    """
    pardict = {}
    with open(parfile,'r+') as f:
        for line in f:
            flags = line[56:65].split(' ')
            try:
                flags = [int(f) for f in flags]
            except:
                continue
            # if we found res pars
            if( all(flags) <= 3 ):
                # if any varied pars
                if( any(flags) > 0 ):
                    # energies are dict keys
                    estring = endf_float_str(float(line[0:11]))
                    pardict[estring] = []
                    pars = [float(line[0+11*i:11+11*i]) for i in range(len(flags))]
                    for i,flag in enumerate(flags):
                        if( flag > 0 ):
                            pardict[estring].append((i,pars[i]))
    return pardict

def update_pardict(pardict,mcpars):
    """
    Update the parameter dict using a pandas Series
    of values taken from a MC run 

    Parameters
    ----------
    pardict : dict
        A dict with keys of base res energies and values
        of varied parameters from a SAMMY operation
    mcpars : Series
        A Pandas Series that has pre-set key names from
        a MC SAMMY run (fitAPI)

    Returns
    -------
    none : None
        No return, only updates input parameter `pardict`

    """
    # figure out which pandas key matches dict keys
    colnames = mcpars.keys().tolist()
    keycounter = 0 # where in pardict key are we?
    prevnamenum = 0
    # run over MC names
    for i,name in enumerate(colnames):
        namenum = float(name.split("_")[1])
        if( namenum!= prevnamenum ):
            keycounter = 0
            prevnamenum = namenum
        for key in pardict:
            # get key-energy to see if it matches MC name
            mult = 1
            tempkey=key
            if key[0]=="-":
                tempkey = key[1:len(key)]
                mult = -1
            keynum = mult*float(tempkey.replace('-','e-').replace('+','e+'))
            percdiff = abs((namenum-keynum)/((namenum+keynum)/2))*100
            # found a matching res
            if( percdiff < 0.001 ):
                parname = name.split('_')[0]
                if parname == "Energy":
                    parInd = 0
                elif parname == "gamma":
                    parInd = 1
                elif parname == "IncChan":
                    parInd = 2
                else:
                    raise ValueError("par type: {} is not implemented".format(parname))
                pardict[key][keycounter] = (parInd,mcpars[name])
                keycounter+=1

def strip_line_num(inpfile,outfile):
    """
    Strip out the line numbers in an ENDF file, which occur after
    column number 75

    Parameters
    ----------
    inpfile : str
        The path and name of the input file
    outfile : str
        The path and name of the output file that will have line
        numbers stripped out
    """
    with open(inpfile,"r+") as f:
        inplines = f.readlines()
    with open(outfile,"w+") as f:
        for line in inplines:
            line = line[0:75] + '\n'
            f.write(line)

def loglog_interp(x,xp,fp):
    """
    Translate x,xp,fp into log-log space, linearly interpolate, and
    return them back in the same space as before

    Parameters
    ----------
    x : array-like
        The new grid onto which we want to interpolate
    xp : array-like
        The old grid from which we interpolate
    fp : array-like
        The "y-values" for xp

    Returns
    -------
    y : array-like
        The "y-values" that we linearly interpolated in log-log space
    """
    lx  = np.log(x)
    lxp = np.log(xp)
    lfp = np.log(fp)
    return np.exp(np.interp(lx,lxp,lfp))






import nuctools.rpixdr_tools as rxt
import numpy as np
from numpy.testing import assert_array_equal,assert_allclose
import os

def test_Rpy_xdr():

    mon_sample_0 = np.array([[200000,300000],[200000,300000],
                             [200000,300000],[200000,300000],
                             [200000,300000],[200000,300000],
                             [200000,300000],[200000,300000],
                             [200000,300000],[200000,300000],
                             [22,55]]) # <- detector sums
    mon_sample_1 = np.array([[200000,300000],[200000,300000],
                             [200000,300000],[200000,300000],
                             [200000,300000],[200000,300000],
                             [200000,300000],[200000,300000],
                             [200000,300000],[200000,300000],
                             [22,55]]) # <- detector sums

    det_sum_0 = np.repeat(2,11) + np.repeat(5,11)
    det_sum_1 = np.repeat(2,11) + np.repeat(5,11)


    # define inputs to the class
    folder = os.getcwd() + '/tests/test_files/'
    number_of_samples = 2

    rpixdr_class = rxt.Rpy_xdr(folder,num_samp=number_of_samples,raw_bw=0.0064)

    assert_array_equal(rpixdr_class.monitors[0],mon_sample_0)
    assert_array_equal(rpixdr_class.monitors[1],mon_sample_1)

def test_Rpy_xdr_std():

    N = 2
    mon = np.array((200000,300000))
    det = np.array((22,55))
    mr = mon/det

    std = np.repeat(np.sqrt(np.sum( (mr-np.mean(mr))**2 )/N)/np.mean(mr), 10)
    std = np.array((std,std))


    # define inputs to the class
    folder = os.getcwd() + '/tests/test_files/'
    number_of_samples = 2

    rpixdr_class = rxt.Rpy_xdr(folder,num_samp=number_of_samples,raw_bw=0.0064)

    assert_array_equal(std,np.array(rpixdr_class.mon_std)[:,:10])






import pandas as pd
import numpy as np
import warnings
import os,sys

__all__ = ['retrieve_tally','modify_sdef']

def retrieve_tally(infile,cells,numbins,skip_extra=0):
    """
    Find the line numbers where tally data starts in mcnp
    output files given the cell numbers being tallied

    Parameters
    ----------
    infile : ascii file
        The mncp output file
    cells : array like
        The cell or surface numbers that are being tallied, 
        if just one, give the single int in an array
    numbins : int
        The number of bins in the tally
    skip_extra : int, optional
        If there are additional comment lines before the 
        tally data, skip this many lines

    Returns
    -------
    ind : numpy array
        An array of the tally data
    """
    file = open(infile).readlines()
    ind = []
    for cell_number in cells:
        for i,line in enumerate(file):
            cell_str = "cell  {} ".format(cell_number)
            surf_str = "surface  {} ".format(cell_number)
            if cell_str in line or surf_str in line:
                ind.append(i+2)

    tally_list = []

    for row in ind:
        temp = np.genfromtxt(infile,skip_header=row+skip_extra,max_rows=numbins,unpack=True)
        tally_list.append(temp)

    return tally_list


def modify_sdef(source,mcnp_file_name,si_string,sp_string):
    """
    Modify the source information and source probability cards in 
    MCNP

    Parameters
    ----------
    source : array-like
        An array of source information and probability. source[0] is
        what will be defined as the "source information" and source[1]
        is what will be defined as the "source probability".
    mcnp_file_name : string
        The full file path including the name of the file
    si_string : str
        The string that exists in the MCNP input file you want to modify,
        this string will be used to identify the line to replace for the 
        source information (gamma energies) Example is \'SI4\'. Only 
        include the card string, it should be unique.
    sp_string : str
        The string that exists in the MCNP input file you want to modify,
        this string will be used to identify the line to replace for the 
        source probability (frequency of gamma energies occuring) Example 
        is \'SP4\'. Only include the card string, it should be unique.

    """
    # ---------------------------------------------------------
    # ----- Write a new input file with the spectrum
    # ---------------------------------------------------------
    found_si_string = False
    found_sp_string = False

    if '/' in mcnp_file_name:
        filesplit = mcnp_file_name.split('/')
    else:
        filesplit = mcnp_file_name.split('\\')
    filesplit[len(filesplit)-1] = 'mod_'+filesplit[len(filesplit)-1]
    newfilename = '/'.join(filesplit)

    source = np.array(source)

    frequencies = source[1]/source[1].sum()

    with open(mcnp_file_name,'r+') as f:
        start_file = f.readlines()
    with open(newfilename,'w+') as f:
        for line in start_file:
            if si_string in line:
                found_si_string = True
                f.write(si_string+' A')
                for info in source[0]:
                    f.write('      {:1.4e}\n'.format(info))
            elif sp_string in line:
                found_sp_string = True
                f.write(sp_string)
                for frequency in frequencies:
                    f.write('      {:1.4e}\n'.format(frequency))
            else:
                f.write(line)

    if found_si_string == False:
        warnings.warn('si_string was never found')
    if found_sp_string == False:
        warnings.warn('sp_string was never found')

    print("File written: {}".format(newfilename))
















from nuctools import *
import numpy as np
import pandas as pd
from numpy.testing import assert_array_equal,assert_allclose
import os

def test_Trans():
	open_file_name = os.getcwd() + '/tests/test_files/test_Ogrp_open.dat'
	samp_file_name = os.getcwd() + '/tests/test_files/test_Ogrp_samp.dat'
	oMon_file_name = os.getcwd() + '/tests/test_files/test_Mon_open.dat'
	sMon_file_name = os.getcwd() + '/tests/test_files/test_Mon_samp.dat'

	odat = pd.read_csv(open_file_name,delim_whitespace=True,
		               names=['tof','bin_width','counts'])
	sdat = pd.read_csv(samp_file_name,delim_whitespace=True,
		               names=['tof','bin_width','counts'])
	omon = pd.read_csv(oMon_file_name,names=['mon'])
	smon = pd.read_csv(sMon_file_name,names=['mon'])

	# default behavior of Trans
	odat['cps'] = odat.counts/odat.bin_width/omon.mon[0]/1e-6
	sdat['cps'] = sdat.counts/sdat.bin_width/smon.mon[0]/1e-6

	odat['bkg'] = np.repeat(0.5,len(odat))/odat.bin_width/omon.mon[0]/1e-6
	sdat['bkg'] = np.repeat(0.5,len(sdat))/sdat.bin_width/smon.mon[0]/1e-6

	trans = ( sdat.cps - sdat.bkg ) / ( odat.cps - odat.bkg )

	Trans_test = Trans(open_file_name,oMon_file_name,
		               samp_file_name,sMon_file_name,
		               o_bkgcoeff=[0.0,0.0,0.5/odat.bin_width[0]/omon.mon[0]/1e-6],
		               s_bkgcoeff=[0.0,0.0,0.5/odat.bin_width[0]/omon.mon[0]/1e-6])

	assert_array_equal(Trans_test.trans,trans)

def test_calc_count_rate():
	"""
	This test may be superfluous
	"""

	counts = 2.0
	bin_width = 0.5
	triggers = 10

	# counts/bin_width/triggers/1e-6
	cps = 0.4/1e-6
	# np.sqrt(counts)/bin_width/triggers/1e-6
	dcps = np.sqrt(counts)/5/1e-6

	assert cps,dcps == Trans.calc_count_rate(counts,bin_width,triggers)

























import pandas as pd
import numpy as np
import scipy.integrate as intgr
import time
from . import physics_tools as pt
from . import sam_tools as st
from . import math_tools as mt

__all__ = ['tofe','etof','opsum','single_group','comp_group','gelina_group','sgfilter',
           'optstat_group','calc_notch']


def tofe(tof,FP):
    """
    Convert neutron time-of-flight into energy

    Parameters
    ----------
    tof: float
        Time of flight [us] of neutron in micro-seconds
    FP: float
        flight path length [m] of neutron in meters

    Returns
    -------
    E: 
        Energy [eV] of neutron that traveled the specified
        flight path in the specified time. (Similar to
        E=0.5mv^2)

    Notes
    -----
    Accounts for relativistic corrections.

    .. math:: E(TOF,FP) = m_n c^2 \\left( \\frac{1}{\\sqrt{1-\\frac{\\left(\\frac{FP}{TOF}\\right)^2}{c^2}}} - 1 \\right)

    Examples
    --------
    >>> import numpy as np
    >>> import nuctools as nuc
    >>> tof = np.linspace(0.1,2500,1e4)
    >>> FP = 100.141 
    >>> E = nuc.tofe(tof,FP)

    """
    #if type(tof) == float or type(tof) == int:
    #    test_tof = np.array([tof])
    #if type(tof) == list:
    #    test_tof = tof
    #    tof = np.array(tof)
    #if type(tof) == np.ndarray:
    #    test_tof = tof
    #    tof = np.array(tof)
    #if True in [FP/x/c > 1 for x in test_tof]:
    #    raise ValueError("FP/TOF/c > 1, input is incorrect.")

    E = pt.mnc2*(1/np.sqrt(1-(FP/tof/pt.c)**2)-1)
    return E
#------------------------------------------------------------------------------
def etof(E,FP):
    """
    Convert neutron energy into time-of-flight

    Parameters
    ----------
    E: float
        Energy [eV] of neutron
    FP: float
        flight path length [m] of neutron in meters

    Returns
    -------
    tof: 
        Time of flight [us] for neutron of energy E to travel
        the specified distance. (Similar to E=0.5mv^2)

    Notes
    -----
    Accounts for relativistic corrections.
    
    .. math:: TOF(E,FP) = \\frac{FP}{c}\\frac{1}{\\sqrt{1-\\frac{1}{(\\frac{E}{m_nc^2} + 1)^2}}}
        
    Examples
    --------
    >>> import numpy as np
    >>> import nuctools as nuc
    >>> E = np.linspace(10,1e6,1e4)
    >>> FP = 100.141 
    >>> E = nuc.etof(E,FP)

    """
    if type(E) == list:
        E = np.array(E)
    tof = FP/pt.c*1/np.sqrt(1-1/(E/pt.mnc2+1)**2)
    return tof
#------------------------------------------------------------------------------
def opsum(infile_list,outfile,numfiles,hlen):
    """
    Function to sum multiple norsum format open files

    Takes norsum grouped style format files and sums them, the 
    header stays the same as first file except summation of
    the triggers from all files. Keeps the same channel number,
    sums the counts, quadrature sum of the error. 

    .. math:: \\sqrt{a^2+b^2}

    Parameters
    ----------
    infile_list: list
        Strings of filenames, e.g. "C:/Users/example.txt"
    outfile: str
        Filename for newly summed data
    numfiles: int
        Number of files to be summed into outfile
    hlen: int
        Header length: how many lines are in the header
    
    Returns
    -------
    None
        Writes newly summed data to text file.

    """
    file_array = []
    # create array of files where each file is vector of line strings
    for infile in infile_list:
        with open(infile,"r+") as file:
            file_array.append(file.readlines())
    print("Number of files: ",len(file_array))
    # write the header and break after triggers line
    with open(outfile,"w+") as newfile:
        breakFlag = False
        tot_trig = 0
        for i,line in enumerate(file_array[0]):
            if breakFlag:
                break
            # sum the triggers from all files
            if "total number of triggers" in line:
                breakFlag = True
                for file in file_array:
                    print(int(file[i].split()[0]))
                    # get the triggers from each file and sum them
                    tot_trig += int(file[i].split()[0])
                line = "{}                      total number of triggers\n".format(tot_trig)
            newfile.write(line)
    # free memory
    del file_array
    data = []
    for file in infile_list:
        print(file)
        data.append(np.genfromtxt(file,unpack=True,skip_header=hlen))
    # The shape of the data should be ( #files x #cols x #rows ) = 3 dims.
    if len(np.shape(data)) < 3:
        raise ValueError("Files do not match. ( # rows, # cols )")
    # Throw an error if the number of columns in the unpacked file is not 3
    if np.shape(data)[1] != 3:
        raise ValueError("Shape of data is unexpected. File format may be incorrect.")
    data = np.array(data)
    # sum the counts across the files
    counts = data.sum(axis=0)[1]
    # quadrature sum of the errors
    err = np.sqrt((data**2).sum(axis=0)[2])
    sum_data = np.vstack((data[0][0],counts,err)).transpose()
    # append summed data to the outfile
    with open(outfile,"ab") as newfile:
        np.savetxt(newfile,sum_data,fmt=["%-15d","%8.5f","%16.10f"])
    # free memory
    del sum_data
    print("done")
#------------------------------------------------------------------------------
def single_group(data,factor,avg=False,binedge=False,prop_err=False):
    """
    Group time of flight data by a single factor

    Grouping is done using the numpy function resize. Each column of data
    is broken into len(data[0])/factor rows, each row is length factor, 
    to find group tof each row of data[0] is averaged (default), each 
    row of data[1] is summed (default).

    Parameters
    ----------
    data : numpy array 
        First column of data should be tof, second 
        column of data should be counts. If propagation
        of error is desired, 3rd column of data should
        be the error, otherwise 3rd column is optional
    factor : int
        how many bins to group by
    avg : bool, optional
        Return an average of group values instead of
        the sum in the group
    binedge : bool, optional
        Return TOF column for grouped data as the first
        tof in group instead of the mean tof of the
        group.
    prop_err : bool, optional
        IF using the avg option, propagate the error given
        in the data array instead of calculating the 
        standard deviation of the mean of each group
    Returns
    -------
     : 2-d numpy array
        Grouped data with columns: tof,bin_width,counts,error counts

    Notes
    -----
    Data array must have more rows than columns, this is 
    typical of TOF data. 

    Examples
    --------
    >>> import nuctools as nuc
    >>> tof = np.linspace(100,1200,400000)
    >>> counts = nuc.noisy_gauss(tof,170,7)
    >>> data = [tof,counts]
    >>> gdata = nuc.single_group(data,2)
    >>> plt.figure(3)
    >>> plt.plot(tof,gauss,color="k",ls=" ",marker="o")
    >>> plt.plot(gdata[0],gdata[1])
    >>> plt.show()

    """
    # data is array with tof & counts for columns
    ld = max(np.shape(data[0]))
    # col. len. should be greatest integer division possible
    col_len = ld//factor
    # if user asked for leading tof bin edge
    if binedge == True:
        # tof for bin is time of first element in bin vector
        gtof = np.resize(data[0],(col_len,factor))[:,0]
    else:
        # tof for bin is mean of tof bin vector
        gtof = np.resize(data[0],(col_len,factor)).mean(axis=1)
    if avg == True:
        # counts for bin is the mean of all elements in bin vector
        gcounts = np.resize(data[1],(col_len,factor)).mean(axis=1)
        if prop_err == True:
            # sqrt of the sum of the squares
            ergcounts = 1/factor*np.sqrt(np.resize(data[2]**2,(col_len,factor)).sum(axis=1))
        else:
            stdcounts = np.resize(data[1],(col_len,factor)).std(axis=1)
            ergcounts = stdcounts/np.sqrt(factor)
    else:
        # counts for bin is the sum of all elements in bin vector
        gcounts = np.resize(data[1],(col_len,factor)).sum(axis=1)
        ergcounts = np.sqrt(gcounts)
    # difference between time bins to plot cps
    dt = np.diff(gtof)
    # diff is calc. between points, therefore has n-1 pts -- add 1 at 0
    dt = np.insert(dt,0,dt[0])
    
    return np.array([gtof,dt,gcounts,ergcounts])
#------------------------------------------------------------------------------
def comp_group(data,factor,c_pts,FP,avg=False,binedge=False,verbose=False):
    """
    Group counts data with compression points as a function of TOF.

    Group TOF data with different factors i.e compression points according
    to the group selected by the compression points. Function finds indices
    of the tof data (data[0]). These tof indices match closely the TOF's given
    in c_pts. Then the function forms len(c_pts)+1 compression regions, and groups each 
    compression region by the values specified by the user in parameter: factor.

    Parameters
    ----------
    data : 2-d numpy array
        Array with columns "tof","counts", and "error".
    factor : 1-d numpy array
        Compression factors for the tof groups, 1st compression factor
        corresponds to time of flights below c_pt[0], 2nd corresponds to tofs above
        c_pt[0] and below c_pt[1], and so on until end of range is found.
    c_pts : 1-d numpy array
        Compression time of flights that seperate the bins that are to be grouped.
        List points from lowest to highest (should re-write to sort and remove user
        capability to fail.)
    FP : float
        Flight path needed to convert the time of flight values of bins to
        energies for the compression point selection.
    avg : bool, optional
        Return an average of group values instead of
        the sum in the group. group error is 1/N*(sum(error^2))^(1/2)
    binedge : bool, optional
        Return a TOF column for grouped data as the first
        tof in group as well as the mean tof of the group.
    verbose : bool, optional
        Choose whether to print out approximate comp pt energy locations
        and indices. Default is False.

    Returns
    -------
    gdata : 2-d numpy array
        Returns the grouped data with columns: tof,bin_width,counts,error counts
    ascii-file : text file (NOT YET)
        Contains a header of bin widths and indices of where groups end, 
        and the grouped data.

    Notes
    -----
    The number of values in tof data, must be greater than 3 (number of columns 
    defined in the description of parameter: data[tof_i,counts_i,errCounts_i])

    The tof vector no longer contains all the information for bin width, it now
    describes the average tof for each grouped bin. This can cause problems at the
    compression points when converting from counts to cps. Use the output bin_width
    to calculate cps.

    Examples
    --------
    >>> import nuctools as nuc
    >>> import matplotlib as plt
    >>> tof = np.linspace(100,1200,1000)
    >>> counts = nuc.noisy_gauss(tof,170,7) + nuc.noisy_gauss(tof,270,15) 
    >>> counts += nuc.noisy_gauss(tof,400,20)
    >>> counts *= 1000
    >>> ercounts = np.sqrt(counts)
    >>> FP = 45.25
    >>> factor,c_pts = np.array([2,4,8]),np.array([100,10])
    >>> data = [tof,counts,ercounts]
    >>> gdata = nuc.comp_group(data,factor,c_pts,FP)
    >>> plt.figure(1)
    >>> plt.plot(tof,counts,color="blue",label="orig")
    >>> plt.plot(gdata[0],gdata[1],label="grouped")
    >>> plt.legend()
    >>> plt.show()

    """
    # cast data as a numpy array
    data = np.array(data)
    # ensure shape is correct (pandas input would've caused probs.)
    data = np.reshape(data,np.sort(np.shape(data)))
    # if statement for compression pts out of tof range
    for val in c_pts:
        if val > max(data[0]) or val < min(data[0]):
            print("Tof: ",max(data[0]),min(data[0]))
            raise ValueError("Compression pt. is out of time-of-flight range.")
    # should be more than one factor
    if len(factor) < 2:
        raise ValueError("Only one factor was given, use single_group func.")
    # cast factor and c_pts to np.array()
    factor,c_pts = np.array(factor),np.array(c_pts)
    if verbose == True:
        print("Comp. pt at ",tofe(c_pts,FP)," eV")
    if len(c_pts)+1 != len(factor):
        raise ValueError("Length of c_pts should be 1 less than length factor")
    # --- Selecting compression regions -----------------------------
    c_ind,j = [0],0
    ng,ld = len(factor),len(data[0])
    # for each compression point go find the index where it lands in the
    # tof grid, and select an indice close to that tof where the grouping
    # can include all bins (except at the end of tof range)
    for i,pt in enumerate(c_pts):
        # index = (tof-first time of flight)/bin_width
        index = int((pt-data[0][0])/(data[0][1]-data[0][0]))
        while (index-c_ind[i]) % factor[j]:
            index += 1
        j += 1
        c_ind.append(index)
        if verbose == True:
            print("compression indices ",index)
    if c_ind[len(c_ind)-1] + 2*factor[len(factor)-1] > ld:
        print('You\'ve grouped too hard on the last compression')
        print('region. You need at least 2 bins for each, and ')
        print('there is only room for one.')
        raise ValueError("Redefine factor for last comp. region.")
    c_ind = c_ind[1:len(c_ind)]
    # ---------------------------------------------------------------
    gdata = []
    # loop over multiplication factors: mult
    for i, mult in enumerate(factor):
        # if in first compression point group
        if i == 0:
            # shape of temporary array (rows,cols)
            # rows are how many groups we'll have, cols are how 
            # many elements are being binned
            shape = (c_ind[i]//mult,mult)
            # tof for bin is mean time in bin vector
            # resize the data and take the average of the row
            #gtof = [vector.mean() for vector in np.resize(data[0][0:c_ind[0]],shape)]
            gtof = np.resize(data[0][0:c_ind[0]],shape).mean(axis=1)
            # if user selected to return tof of leading bin edge
            if binedge == True:
                # tof for bin is time of first element in bin vector
                edge_gtof = np.resize(data[0][0:c_ind[0]],shape)[:,0]
            # calculate the average counts in data[1]
            if avg == True:
                gcounts = np.resize(data[1][0:c_ind[0]],shape).mean(axis=1)
                ergcounts = 1/mult*np.sqrt(np.resize(data[2][0:c_ind[0]]**2,shape).sum(axis=1))
            # calculate the sum of counts in data[1]
            else:
                # counts for bin is the sum of all elements in bin vector
                # resize the data and take the sum of the row
                gcounts = np.resize(data[1][0:c_ind[0]],shape).sum(axis=1)
                ergcounts = np.sqrt(gcounts)
            # difference between time bins to plot cps
            dt = np.diff(gtof)
            # diff is calc. between points, therefore has n-1 pts -- add 1 at 0
            dt = np.insert(dt,0,dt[0])
            # if user asked for leading tof bin edge append it as an additional 
            # column to gdata
            if binedge == True:
                gdata.append([gtof,dt,gcounts,ergcounts,edge_gtof])
            else:
                gdata.append([gtof,dt,gcounts,ergcounts])
            continue
        # if in the last compression group (in tof)
        if i == ng-1:
            # shape of temporary array
            shape = ((ld-c_ind[i-1])//mult,mult)
            # tof for bin is mean time in bin vector
            gtof = np.resize(data[0][c_ind[i-1]:ld],shape).mean(axis=1)
            # if user selected to return tof of leading bin edge
            if binedge == True:
                # tof for bin is time of first element in bin vector
                edge_gtof = np.resize(data[0][c_ind[i-1]:ld],shape)[:,0]
            # calculate the average counts in data[1]
            if avg == True:
                gcounts = np.resize(data[1][c_ind[i-1]:ld],shape).mean(axis=1)
                ergcounts = 1/mult*np.sqrt(np.resize(data[2][c_ind[i-1]:ld]**2,shape).sum(axis=1))
            # calculate the sum of counts in data[1]
            else:
                # counts for bin is the sum of all elements in bin vector
                gcounts = np.resize(data[1][c_ind[i-1]:ld],shape).sum(axis=1)
                ergcounts = np.sqrt(gcounts)
            # difference between time bins to plot cps
            dt = np.diff(gtof)
            # diff is calc. between points, therefore has n-1 pts -- add 1 at 0
            dt = np.insert(dt,0,dt[0])
            # if user asked for leading tof bin edge append it as an additional 
            # column to gdata
            if binedge == True:
                gdata.append([gtof,dt,gcounts,ergcounts,edge_gtof])
            else:
                gdata.append([gtof,dt,gcounts,ergcounts])
            continue
        else:
            # shape of temporary array
            shape = (((c_ind[i]-(c_ind[i-1]))//mult),mult)
            # tof for bin is avg of elements in bin vector
            # resize the data and take the average of the row
            gtof = np.resize(data[0][c_ind[i-1]:c_ind[i]],shape).mean(axis=1)
            # if user selected to return tof of leading bin edge
            if binedge == True:
                # tof for bin is time of first element in bin vector
                edge_gtof = np.resize(data[0][c_ind[i-1]:c_ind[i]],shape)[:,0]
            # calculate the average counts in data[1]
            if avg == True:
                gcounts = np.resize(data[1][c_ind[i-1]:c_ind[i]],shape).mean(axis=1)
                ergcounts = 1/mult*np.sqrt(np.resize(data[2][c_ind[i-1]:c_ind[i]]**2,shape).sum(axis=1))
            # calculate the sum of counts in data[1]
            else:
                # counts for bin is the sum of all elements in bin vector
                # resize the data and take the sum of the row
                gcounts = np.resize(data[1][c_ind[i-1]:c_ind[i]],shape).sum(axis=1)
                ergcounts = np.sqrt(gcounts)
            # difference between time bins to plot cps
            dt = np.diff(gtof)
            # diff is calc. between points, therefore has n-1 pts -- add 1 at 0
            dt = np.insert(dt,0,dt[0])
            # if user asked for leading tof bin edge append it as an additional 
            # column to gdata
            if binedge == True:
                gdata.append([gtof,dt,gcounts,ergcounts,edge_gtof])
            else:
                gdata.append([gtof,dt,gcounts,ergcounts])
            
    # hstack in this case is appending all the gdata arrays into one array 
    # with columns gtof,gcounts,ergcounts
    gdata = np.hstack([val for val in gdata])
        
    return gdata
#------------------------------------------------------------------------------
def gelina_group(data,factor,c_ind,avg=False,binedge=False,verbose=False):
    """
    Group like AGL for GELINA comparison. This is experimental for now.

    Parameters
    ----------
    data : 2-d numpy array
        Array with columns "tof","counts", and "error".
    factor : 1-d numpy array
        Compression factors for the tof groups, 1st compression factor
        corresponds to time of flights below c_pt[0], 2nd corresponds to tofs above
        c_pt[0] and below c_pt[1], and so on until end of range is found.
    c_pts : 1-d numpy array
        Compression points that define the edges of the bins that are to be grouped.
        List points from lowest to highest (should re-write to sort and remove user
        capability to fail.)
    verbose : bool
        Whether to print out information to STDOUT

    Returns
    -------
    gdata : 2-d numpy array
        Returns the grouped data with columns: tof,bin_width,counts,error counts

    """
    ng,ld = len(factor),len(data[0])
    gdata = []
    # loop over multiplication factors: mult
    for i, mult in enumerate(factor):
        # if in first compression point group
        if i == 0:
            # shape of temporary array (rows,cols)
            # rows are how many groups we'll have, cols are how 
            # many elements are being binned
            shape = (c_ind[i]//mult,mult)
            # tof for bin is mean time in bin vector
            # resize the data and take the average of the row
            #gtof = [vector.mean() for vector in np.resize(data[0][0:c_ind[0]],shape)]
            gtof = np.resize(data[0][0:c_ind[0]],shape).mean(axis=1)
            # if user selected to return tof of leading bin edge
            if binedge == True:
                # tof for bin is time of first element in bin vector
                edge_gtof = np.resize(data[0][0:c_ind[0]],shape)[:,0]
            # calculate the average counts in data[1]
            if avg == True:
                gcounts = np.resize(data[1][0:c_ind[0]],shape).mean(axis=1)
                ergcounts = 1/mult*np.sqrt(np.resize(data[2][0:c_ind[0]]**2,shape).sum(axis=1))
            # calculate the sum of counts in data[1]
            else:
                # counts for bin is the sum of all elements in bin vector
                # resize the data and take the sum of the row
                gcounts = np.resize(data[1][0:c_ind[0]],shape).sum(axis=1)
                ergcounts = np.sqrt(gcounts)
            # difference between time bins to plot cps
            dt = np.diff(gtof)
            # diff is calc. between points, therefore has n-1 pts -- add 1 at 0
            dt = np.insert(dt,0,dt[0])
            # if user asked for leading tof bin edge append it as an additional 
            # column to gdata
            if binedge == True:
                gdata.append([gtof,dt,gcounts,ergcounts,edge_gtof])
            else:
                gdata.append([gtof,dt,gcounts,ergcounts])
            continue
        # if in the last compression group (in tof)
        if i == ng-1:
            # shape of temporary array
            shape = ((ld-c_ind[i-1])//mult,mult)
            # tof for bin is mean time in bin vector
            gtof = np.resize(data[0][c_ind[i-1]:ld],shape).mean(axis=1)
            # if user selected to return tof of leading bin edge
            if binedge == True:
                # tof for bin is time of first element in bin vector
                edge_gtof = np.resize(data[0][c_ind[i-1]:ld],shape)[:,0]
            # calculate the average counts in data[1]
            if avg == True:
                gcounts = np.resize(data[1][c_ind[i-1]:ld],shape).mean(axis=1)
                ergcounts = 1/mult*np.sqrt(np.resize(data[2][c_ind[i-1]:ld]**2,shape).sum(axis=1))
            # calculate the sum of counts in data[1]
            else:
                # counts for bin is the sum of all elements in bin vector
                gcounts = np.resize(data[1][c_ind[i-1]:ld],shape).sum(axis=1)
                ergcounts = np.sqrt(gcounts)
            # difference between time bins to plot cps
            dt = np.diff(gtof)
            # diff is calc. between points, therefore has n-1 pts -- add 1 at 0
            dt = np.insert(dt,0,dt[0])
            # if user asked for leading tof bin edge append it as an additional 
            # column to gdata
            if binedge == True:
                gdata.append([gtof,dt,gcounts,ergcounts,edge_gtof])
            else:
                gdata.append([gtof,dt,gcounts,ergcounts])
            continue
        else:
            # shape of temporary array
            shape = (((c_ind[i]-(c_ind[i-1]))//mult),mult)
            # tof for bin is avg of elements in bin vector
            # resize the data and take the average of the row
            gtof = np.resize(data[0][c_ind[i-1]:c_ind[i]],shape).mean(axis=1)
            # if user selected to return tof of leading bin edge
            if binedge == True:
                # tof for bin is time of first element in bin vector
                edge_gtof = np.resize(data[0][c_ind[i-1]:c_ind[i]],shape)[:,0]
            # calculate the average counts in data[1]
            if avg == True:
                gcounts = np.resize(data[1][c_ind[i-1]:c_ind[i]],shape).mean(axis=1)
                ergcounts = 1/mult*np.sqrt(np.resize(data[2][c_ind[i-1]:c_ind[i]]**2,shape).sum(axis=1))
            # calculate the sum of counts in data[1]
            else:
                # counts for bin is the sum of all elements in bin vector
                # resize the data and take the sum of the row
                gcounts = np.resize(data[1][c_ind[i-1]:c_ind[i]],shape).sum(axis=1)
                ergcounts = np.sqrt(gcounts)
            # difference between time bins to plot cps
            dt = np.diff(gtof)
            # diff is calc. between points, therefore has n-1 pts -- add 1 at 0
            dt = np.insert(dt,0,dt[0])
            # if user asked for leading tof bin edge append it as an additional 
            # column to gdata
            if binedge == True:
                gdata.append([gtof,dt,gcounts,ergcounts,edge_gtof])
            else:
                gdata.append([gtof,dt,gcounts,ergcounts])
            
    # hstack in this case is appending all the gdata arrays into one array 
    # with columns gtof,gcounts,ergcounts
    gdata = np.hstack([val for val in gdata])
        
    return gdata


#------------------------------------------------------------------------------
def optstat_group(data,FP,D,res_per_bin,E_cpts=[1e2,1e3,1e4],verbose=False,
                  t0=0.0,binedge=False):
    """
    Function that finds the grouping necessary to include at least res_per_bin
    resonances in each group.

    Finds the grouping scheme necessary to include at least res_per_bin 
    resonances in each group and then groups the provided data using the 
    nuctools comp_group function. The E_cpts are estimated compression pts
    for the grouper. This is typically helpful in the URR when you want to 
    group with a \"statiscally significant\" number of resonances.

    Parameters
    ----------
    data : array-like
        Data array. Should be a tof vector at data[0], data[0] must
        have equal width time bins. data[1] should be counts, data[1]
        will be grouped and summed.
    FP : float
        Flight path of the experiment, this helps in determining the
        energy to time ratio
    D : float
        nuclear energy level / resonance spacing
    res_per_bin : int
        The number of resonances per bin that are desired
    E_cpts : array-like, optional
        The compression pts desired, given in energy
    verbose : bool, optional
        Whether to spit out extra information about the grouping
    t0 : float
        Use t0 to determine SAMMY input printout
    binedge : bool, optional
        Return a TOF column for grouped data as the first
        tof in group as well as the mean tof of the group.

    Returns
    -------
    gdata : numpy array
        Returns the grouped data with columns: tof,bin_width,counts,error counts
    """
    # min size in energy
    min_bin_size = D*res_per_bin
    # tof compression points
    tof_pts = etof(E_cpts,FP)
    # bin width in time
    bw = data[0][1]-data[0][0]
    
    data_region_start_pts = np.hstack((tof_pts[::-1],data[0][len(data[0])-1]))
    data_region_start_pts_E = tofe(data_region_start_pts,FP)
    # min size in time
    min_tof_bin = data_region_start_pts-etof(data_region_start_pts_E+min_bin_size,FP)
    # what factor is required for min size
    g_factors = np.array(min_tof_bin/bw).astype(int)
    # don't group lowest TOF values so hard
    g_factors[len(g_factors)-1] = g_factors[len(g_factors)-2]
    # group highest TOF values 3 times as hard
    g_factors[0] = g_factors[0]*3
    if verbose:
        print("E cpts:",data_region_start_pts_E)
        print("Min. bin size [us]:",min_tof_bin)
        print("Grouping factors:",g_factors)
        print("TOF comp. pts:",tof_pts[::-1])
        print("--------------")
        st.print_chann(tof_pts[::-1],g_factors,bw*1e3,t0,FP)
    
    gdata = comp_group(data,g_factors,tof_pts[::-1],FP,binedge=binedge)
    
    return gdata
#------------------------------------------------------------------------------
def sgfilter(f,df,window,poly,deriv=0):
    """
    Performs a Savitsky-Golay filter on data, returning the
    smoothed data and the newly weighted errors on the 
    smoothed data.
    
    Parameters
    ----------
    f : array_like
        The function or data to be smoothed
    df : array_like
        The error associated with f
    window : int
        The window over which to weight the data
    poly : int
        The order of polynomial to used to fit the data
    deriv : int, optional
        The order of derivative to compute. Must be non-
        negative. Default is 0.
    
    Returns
    -------
    s : ndarray
        2d array that has smoothed data in column 0, and
        the error weighted by the smoothing in column 1.

    Notes
    -----
    
    The design matrix :math:`\\mathbf{A}` for a smoothing window of 5 points and a polynomial fit of 
    order 2 is:

    .. math:: \\mathbf{A} = \\left[\\begin{array}{ccc} -2^0 & -2^1 & -2^2 \\cr -1^0 & -1^1 & -1^2 \\cr  0^0 &  0^1 &  0^2 \\cr  1^0 &  1^1 &  1^2 \\cr  2^0 &  2^1 &  2^2 \\cr \\end{array}\\right]

    Or for a smoothing window of length :math:`w`, and polynomial of order :math:`M` where 
    :math:`nL = -w\\div2`, (note that :math:`\\div` denotes integer division) and :math:`nR = nL+w`:

    .. math:: \\mathbf{A}= \\left[\\begin{array}{ccc} -nL^0 & ...   & -nL^M \\cr \\vdots & \\vdots & \\vdots \\cr 0^0   & ...   & 0^M   \\cr \\vdots & \\vdots & \\vdots \\cr nR^0  & ...   & nR^M  \\cr \\end{array}\\right]

    The Savitsky-Golay Coefficients :math:`\\mathbf{C}_{sv}` are given by:

    .. math:: \\mathbf{C}_{sv} = \\left(\\mathbf{A}^T\\mathbf{A}\\right)^{-1}\\mathbf{A}^T

    The smoothed signal :math:`Y_{i,j}` comes from the convolution (:math:`\\circledast`) of the original signal :math:`y_i` (:math:`i\\in0...N`) with each column :math:`j` in the matrix :math:`\\mathbf{C}_{sv}`:

    .. math:: Y(t)_j = \\mathbf{C}_{sv,j}\\circledast y_i = \\sum_{k=-nL}^{nR} \\mathbf{C}_{sv,k+nL} \\cdot y_{i+k}

    for (i :math:`\\in nL...N-nR-1`)

    Examples
    --------
    >>> import nuctools as nuc
    >>> import matplotlib.pyplot as plt
    >>> tof = np.linspace(100,250,150)
    >>> counts = nuc.noisy_gauss(tof,170,7)
    >>> counts *= 1000
    >>> ercounts = np.sqrt(counts)
    >>> sm_counts = nuc.sgfilter(counts,ercounts,5,0)
    >>> plt.figure()
    >>> plt.errorbar(x,counts,ercounts,label="orig. data")
    >>> plt.errorbar(x,sm_counts[0],sm_counts[1],label="sm. data")
    >>> plt.legend()
    >>> plt.show()

    .. image:: images/sav_gol_example.png

    """
    
    
    if poly >= window:
        raise ValueError("polyorder must be less than window_length.")

    pos, rem = divmod(window, 2)

    if rem == 0:
        raise ValueError("window_length must be odd.")
    
    nL,nR = pos,window-pos
    
    # The design matrix A
    A = np.array([[i**j for i in range(-nL,nR)] for j in range(poly+1)])
    B = np.linalg.inv(np.dot(A,np.transpose(A)))
    # The Savitsky-Golay coefficients
    SV = np.transpose(np.dot(np.transpose(A),B))

    lf = len(f)
    smooth,sigsmooth = np.array(f),np.array(df)**2
    # ------ leaving this here in case speedy version is confusing -----------
    # for each point to smooth
    #for i in range(nL,lf-nR-1):
        #smooth[i] = 0
        # for each point on the window
        #smooth[i] = (SV[deriv][0:nL+nR]*f[i-nL:i+nR]).sum()
        #sigsmooth[i] = ((SV[deriv][0:nL+nR]*df[i-nL:i+nR])**2).sum()
    smooth[nL:lf-nR+1] = np.convolve(SV[deriv],f,mode='valid')
    sigsmooth[nL:lf-nR+1] = np.convolve(SV[deriv]**2,df**2,mode='valid')
    # ------------------------------------------------------------------------

    dsmooth = np.sqrt(sigsmooth)

    s = np.vstack([smooth,dsmooth])
    
    return s

#------------------------------------------------------------------------------
def calc_notch(spect,lim1,lim2):
    """
    Calculate the average count rate at time-of-flight (TOF) values
    that are within the limits lim1, lim2

    Parameters
    ----------
    spect : Pandas DataFrame
        Must have columns of: "tof","cps","dcps"
    lim1 : float
        lower TOF value limiting average
    lim2 : float
        higher TOF value limiting average

    Returns
    -------
    notch : array
        Numpy array with 3 values: avg TOF, avg count rate, error on avg count rate
    """
    condition = (spect.tof>lim1) & (spect.tof<lim2)
    avg = np.mean(spect.cps[condition])
    avgerr = mt.prop_err_mean(spect.dcps[condition])
    avgtof = np.mean(spect.tof[condition])

    notch = np.array([avgtof,avg,avgerr])
    return notch












import numpy as np
import h5py as h5
import pandas as pd
import time
import warnings

__all__ = ['check_hdf5_triggers','read_odfH5']


def check_hdf5_triggers(hdf5_folder,sample_name):
    """
    Interrogate a standard format RPI linac HDF5 file to compare the 
    triggers recorded by the scaler and the TDC events in the data 
    collection

    Parameters
    ----------
    hdf5_folder : str
        The full filepath where the file lives. Example: "/C:/Data/08Aug17/"
        Should end with a slash
    sample_name : str
        The name sample name associated with the file. This should be in format:
        {name}_master.h5, and *inside* the file, it should start with heading
        name. for example: B4C/CYCLE0001/MON/ and B4C/CYCLE0001/TRIG/ should be 
        paths within the file.

    Returns
    -------
    no return : 
        print statement

    Notes
    -----
    This function is highly dependent on the structure of the datasets in the 
    HDF5 file. This means that users should be very careful when applying this
    to their own data.

    """

    # Warning should be left in until this has been solved.
    warnings.warn("\n\nThis function is requires rigid dataset structure. See notes.\n\n")

    with h5.File("{}{}_master.h5".format(hdf5_folder,sample_name,'r')) as hdf:
        # set stuff to be saved -----------------------------------------------
        percent_diff_array = np.zeros(len(hdf["{}".format(sample_name)]))
        bad_cycles = []
        # ---------------------------------------------------------------------
        for i,cycle in enumerate(hdf["{}".format(sample_name)]):
            mon = np.array(hdf["{}/{}/MON".format(sample_name,cycle)])
            trig = np.array(hdf["{}/{}/TRIG".format(sample_name,cycle)])
            expected_trig = trig[0][0]
            percent_diff = abs(mon[7]-trig[0][1])/((mon[7]+trig[0][1])/2)*100
            # should the percent diff be recorded? ----------------------------
            condition1 = (percent_diff > 100)
            condition2 = (float(trig[0][1])<expected_trig/2)
            condition3 = (float(mon[7])<expected_trig/2)

            if (condition1 or condition2 or condition3):
                print("WARNING: Large trigger difference detected in CYCLE",i)
                percent_diff = 0
                bad_cycles.append('CYCLE '+str(i))
            # print out for the user ------------------------------------------
            print(i+1,"Scal/TDC trigs ===>",mon[7],"/",trig[0][1],
                  "  % Diff: {:.4f}".format(percent_diff))
            percent_diff_array[i] = percent_diff
    
        print("----------------------------")
        print("AVG perc. diff.: {:.4f}%".format(percent_diff_array.mean()))
        print("STD of perc. diff.: {:.4f}".format(percent_diff_array.std()))
        print("----------------------------")
        print("CYCLES that should probably be thrown out (set to 0%):")
        for bad_cycle in bad_cycles:
            print(bad_cycle)
        print("----------------------------")

def read_odfH5(filename):
    """
    Read the h5 file produced by SAMMY conversion from ODF to
    HDF5 format. 

    Parameters
    ----------
    filename : str
        The full file path to the HDF5 file

    Returns
    -------
    data,header : DataFrame tuple
        The experimental data in `data`, and the header info from 
        the ODF file in `header`
    """
    with h5.File(filename,'r') as f:
        # print(list(f.keys()))
        data = f['data/sections'][:]
        colnames = f['data/sectnames'][:]
        header = f['meta/header'][:]
    data = pd.DataFrame(data.T,columns=colnames.astype("U13"))
    header = pd.DataFrame(header)
    header['key'] = header['key'].astype("U13")
    return data,header






import numpy as np
import nuctools.physics_tools as pt

def test_spin_stat_factor():
    J = 2.0
    I = 3.5

    g = (2*J+1)/(2*(2*I+1))

    assert g == pt.spin_stat_factor(J,I)


def test_r_inf():
    r_prime = 9.0
    a_c = 10.0
    Rinf = 1.0 - r_prime/a_c

    r_inf = pt.r_inf(r_prime,a_c=a_c)

    assert Rinf == r_inf

def test_r_prime():
    a = 10.0
    Rinf = 0.1
    Rp_answer = 9.0

    Rp = pt.r_prime(Rinf,a_c=a)

    assert Rp == Rp_answer

def test_wave_number():
    A = 2.0
    E = 2.0

    per_Fermi = 1e-15 # how many meters per Fermi

    # sqrt{2mE}/hbar * A/(A+1) 
    k = 2*np.sqrt(pt.mnc2)/(pt.c*1e6)/pt.hbar * 2/3 * per_Fermi

    assert k == pt.wave_number(A,E)


def test_rho():
    A = 2.0
    E = 2.0
    a_c = 5.0

    per_Fermi = 1e-15 # how many meters per Fermi

    # sqrt{2mE}/hbar * A/(A+1) 
    k = 2*np.sqrt(pt.mnc2)/(pt.c*1e6)/pt.hbar * 2/3 * per_Fermi

    p = k*a_c

    assert p == pt.rho(a_c,A,E)

def test_penetrability():
    A = 2.0
    E = 2.0
    a_c = 5.0

    per_Fermi = 1e-15 # how many meters per Fermi

    # sqrt{2mE}/hbar * A/(A+1) 
    k = 2*np.sqrt(pt.mnc2)/(pt.c*1e6)/pt.hbar * 2/3 * per_Fermi

    p = k*a_c 

    assert pt.penetrability(0,a_c,A,E) == p
    
    assert pt.penetrability(1,a_c,A,E) == p**3/(1+p**2)
    
    assert pt.penetrability(2,a_c,A,E) == p**5/(9+3*p**2+p**4)
    
    assert pt.penetrability(3,a_c,A,E) == p**7/(225+45*p**2+6*p**4+p**6)
    
    assert pt.penetrability(4,a_c,A,E) == p**9/(11025+1575*p**2+135*p**4+10*p**6+p**8)

def test_wigner_spacing():
    x = np.array([2.0])  # spacing
    D = 2.0              # average spacing

    assert pt.wigner_spacing(x,D) == np.pi/4*np.exp(-np.pi/4)




import pandas as pd
import numpy as np
import xml.etree.ElementTree as ET
import h5py as h5
import os,sys,json
import matplotlib.pyplot as plt
import pkgutil

from . import chem_tools as ct
from . import physics_tools as pt

__all__ = ['plot_h5scale_xs',"get_cross_section","get_std_comp","get_std_comp_nat_abund","get_zlist",
           "get_zaidlist","calc_num_densities","write_tsl_table","get_scaleza_name_thermal",
           "get_scaleza_name_metastable","get_scaleza_name_specialNuclei","append_xml_to_table",
           "get_xml_root","get_fastmat_thermal","get_single_mat"]

def plot_h5scale_xs(filename,scaleid,temp,emin=2.1e7,mt=None):
    """
    Plot xs of every MT (reaction number) in SCALE data H5 format

    Parameters
    ----------
    filename : str
        The path to the H5 library file
    scaleid : int
        The integer representing a single isotope or compound, e.g. 1001 
        for hydrogen, 5008016 for the oxygen in BeO 
    temp : float
        The temperature, assumed to only need one digit after the decimal
    emin : float, optional
        Minimum energy of XS must be below this value to be plotted. 
    mt : int, optional
        The MT ENDF/SCALE reaction number to be plotted. Default is all within 
        energy limit defined

    Returns
    -------
    None

    Examples
    --------
    >>> import nuctools as nuc
    >>> 
    >>> scaleid = 5008016
    >>> temperature = 293.6
    >>> nuc.plot_all_h5scale_xs('n_008016.h5',temperature,scaleid)
    """
    with h5.File(filename,'r+') as f:
        plt.figure()
        if mt is None:
            for key in f['n_{:0>7d}_{:0>6.1f}'.format(scaleid,temp)]:
                try:
                    e  = f['n_{:0>7d}_{:0>6.1f}/{}/energy'.format(scaleid,temp,key)][()]
                    xs = f['n_{:0>7d}_{:0>6.1f}/{}/xs'.format(scaleid,temp,key)][()]
                    if e.min() < emin:
                        plt.plot(e,xs,label='{}'.format(key))
                except:
                    print("key = '{}' cannot be plotted".format(key))
        else:
            try:
                e  = f['n_{:0>7d}_{:0>6.1f}/mt_{:0>4d}/energy'.format(scaleid,temp,mt)][()]
                xs = f['n_{:0>7d}_{:0>6.1f}/mt_{:0>4d}/xs'.format(scaleid,temp,mt)][()]
                if e.min() < emin:
                    plt.plot(e,xs,label='{}'.format("mt={}".format(mt)))
            except:
                print("key = 'mt_{:0>4d}' cannot be plotted".format(mt))
        plt.xlabel('Energy [eV]')
        plt.ylabel("Cross section [b]")
        plt.xscale('log')
        plt.yscale('log')
        plt.legend(ncol=3)

def get_cross_section(filename,scaleid,temp,mt):
    """
    Plot xs of every MT (reaction number) in SCALE data H5 format

    Parameters
    ----------
    filename : str
        The path to the H5 library file
    scaleid : int
        The integer representing a single isotope or compound, e.g. 1001 
        for hydrogen, 5008016 for the oxygen in BeO 
    temp : float
        The temperature, assumed to only need one digit after the decimal
    emin : float, optional
        Minimum energy of XS must be below this value to be plotted. 
    mt : int, optional
        The MT ENDF/SCALE reaction number to be plotted. Default is all within 
        energy limit defined

    Returns
    -------
    data : DataFrame
        A Pandas DataFrame with keys: "e" and "cs" for energy and cross section

    Examples
    --------
    >>> import numpy as np
    >>> 
    >>> scaleid = 5008016
    >>> temperature = 293.6
    >>> data = nuc.get_cross_section('n_008016.h5',temperature,scaleid)

    """
    with h5.File(filename,'r') as f:
        try:
            data = pd.DataFrame({
                "e" : f['n_{:0>7d}_{:0>6.1f}/mt_{:0>4d}/energy'.format(scaleid,temp,mt)][()],
                "cs" : f['n_{:0>7d}_{:0>6.1f}/mt_{:0>4d}/xs'.format(scaleid,temp,mt)][()]
            })
        except:
            print("key = 'n_{:0>7d}_{:0>6.1f}/mt_{:0>4d}' cannot be found".format(scaleid,temp,mt))
            df = pd.DataFrame(f['nuclide_md'][:])
            print("Temperatures available for SCALEID = {}:".format(scaleid))
            print(df[df['zaid']==scaleid]['temperature'])
            raise ValueError("Temperature was not found.")
    return data



def get_std_comp(datadir):
    """
    Return a std composition table pandas DataFrame

    assumes the directory and file names are static in the 
    SCALE data directory

    Parameters
    ----------
    datadir : str
        The directory where the "scale data" exists. This is a 
        specific Git repo

    Returns
    -------
    stdcomp : DataFrame
        A table of the standard composition from SCALE data
    """

    std_composition = datadir+'library_inputs/compoz.inp'
    begin = 0
    max_rows = 3527 # num rows in table not including mixtures
    with open(std_composition,'r+') as f:
        for i,line in enumerate(f):
            if "SCID          ROTH.    ICP      NCZA     ATPM     END" in line:
                begin_natural = i+1
    stdcomp = pd.read_fwf(std_composition,skiprows=begin_natural,nrows=max_rows,
                          colspecs=( (0,14), (15,22), (23,27), (29,39), (40,48), (49,55)),
                          names=['name','rho_theor','compound_flag','scaleid','wtpct','end'])
    stdcomp.fillna(0,inplace=True)
    stdcomp.scaleid = stdcomp.scaleid.astype('int')

    return stdcomp

def get_std_comp_nat_abund(datadir):
    """
    Return a std composition table pandas DataFrame for the 
    natural abundance table

    assumes the directory and file names are static in the 
    SCALE data directory

    Parameters
    ----------
    datadir : str
        The directory where the "scale data" exists. This is a 
        specific Git repo

    Returns
    -------
    stdcomp : DataFrame
        A table of the standard composition of natural abundance
        from SCALE data
    """

    std_composition = datadir+'library_inputs/compoz.inp'
    begin_natural = 0
    max_rows = 361 # num rows with natural abundance
    with open(std_composition,'r+') as f:
        for i,line in enumerate(f):
            if "NZN         ISZA   ABWP      END" in line:
                begin_natural = i+1
    stdcomp = pd.read_fwf(std_composition,skiprows=begin_natural,nrows=max_rows,names=['natid','isoid','frac','end'])
    stdcomp.fillna(0,inplace=True)
    stdcomp.isoid = stdcomp.isoid.astype('int')

    return stdcomp

def get_zlist(symbol_list):
    """
    Get list of proton numbers matching the list of symbols given
    in the input

    Parameters
    ----------
    symbol_list : list
        A list of strings with elemental symbols (properly caps), 
        e.g. Am, H, Ca, etc.
    
    Returns
    -------
    zlist : list
        A list of ints for the number of protons in each element
    """
    zlist = []
    f_protons = os.path.join(os.path.dirname(__file__),'data','protons.json')
    with open(f_protons,'r') as f:
        protons = json.load(f)
    for symbol in symbol_list:
        zlist.append(protons[symbol])
    return zlist

def get_zaidlist(zlist,stdcomp):
    """
    Get the list of ZAIDs, relevant to ENDF/SCALE ID system, using the
    standard composition to get the natural composition of isotopes

    Parameters
    ----------
    zlist : list
        List of atomic numbers for unique elements in the salt mixture
    stdcomp : DataFrame
        DataFrame object made exactly like the output of `get_std_comp_nat_abund`

    Returns
    -------
    zaidlist : list
        The list of ZAIDs for the mixture of salts
    """
    zaidlist = []
    natid = 0.0
    for z in zlist:
        for i,za in enumerate(stdcomp.isoid):
            # determine if TSL or metastable (skipping metas!)
            if stdcomp.natid.iloc[i] != 0:
                natid = stdcomp.natid.iloc[i]
            if( z == za//1000 and natid < 1000000):
                zaidlist.append(za)
    return zaidlist

def calc_num_densities(mol_df,mass,area,scaledatadir):
    """
    Calculate number densities for every SCALE ZAID based on 
    mass fractions of molecules input, mass of sample, and 
    area of sample

    Parameters
    ----------
    mol_df : DataFrame
        A Pandas DataFrame with column names: "molecule", and "mass_frac"
        which describe the molecules (with proper capitalization) in the
        sample and the mass fractions of each molecule. 
    mass : float
        The mass of the sample in grams
    area : float
        The area of the sample perpendicular to the beam (neutrons) in cm
    scaledatadir : str
        The path to the SCALE data directory. (This dependency could be 
        removed by including a natural abundance file in nuctools in 
        future releases)

    Returns
    -------
    numdensity_iso : dict
        A dictionary with keys of ZAIDs and values of number density for 
        every ZAID
    
    Example
    -------
    >>> import nuctools as nuc
    >>> import pandas as pd
    >>> mol = pd.DataFrame({"CaO":0.5,"Fe2O3":0.5})
    >>> mass = 40 # grams
    >>> area = 4.5 # cm
    >>> n = nuc.calc_num_densities(mol,mass,area)
    """
    f_protons = os.path.join(os.path.dirname(__file__),'data','protons.json')
    f_unary_list = os.path.join(os.path.dirname(__file__),'data','unary-list.json')
    with open(f_unary_list,'r') as f:
        unary_dict = json.load(f)
        unary_stoich = unary_dict['unary_stoich']
        unary_MM = unary_dict['unary_mol_mass']
    with open(f_protons,'r') as f:
        protons = json.load(f)
    stdcomp = get_std_comp_nat_abund(scaledatadir)

    # get unique elements in sample
    full_elem_list = ct.get_single_elem_list(mol_df.molecule)
    # get proton numbers for all elements in sample
    zlist = [protons[el] for el in full_elem_list]
    # get unique naturally abundant ZAIDs for the elements in sample
    zaidlist = get_zaidlist(zlist,stdcomp)
    # initialize number density dict with zeros
    numdensity_iso = {}
    for zaid in zaidlist:
        numdensity_iso[zaid] = 0.0

    # --- molecule loop
    for i,molec in enumerate(mol_df.molecule):
        MF_mol = mol_df.mass_frac.iloc[i]
        MM_mol = unary_MM[molec]
        elem_list = ct.get_single_elem_list([molec])
        molec_stoich = unary_stoich[molec]
        # --- elements in molecule loop
        for j,elem in enumerate(elem_list):
            z = protons[elem]
            s_elem = molec_stoich[j]
            elem_zaidlist = get_zaidlist([z],stdcomp)
            # --- isotopes in element loop
            for zaid in elem_zaidlist:
                gamma_iso = 0.01 * stdcomp.loc[stdcomp.isoid==zaid].frac.iloc[0]
                numdensity_iso[zaid] += gamma_iso * s_elem * MF_mol * mass * pt.Na / area / MM_mol
    return numdensity_iso

def write_tsl_table(library_master_file,output_file,stdcomp):
    """
    Read library master file to find TSLs and temperatures
    available and write to file

    Parameters
    ----------
    library_master_file : str
        The path for the library master file in HDF5
    output_file : str
        The path where to write a markdown table
    stdcomp : DataFrame
        The SCALE standard composition information which 
        can be placed in a DataFrame with `get_std_comp`

    Returns
    -------
    None


    """
    tsl_table = [['scaleid','name','temps','**notes**']]
    with h5.File(library_master_file,'r') as f:
        last_scaleid = None
        last_scaleid_pos = 0
        for i,nuclide in enumerate(f):
            if nuclide == 'metadata' or nuclide == "nuclide_md" or "p_" in nuclide:
                continue
            found_tsl = False
            scaleid = nuclide.split('_')[1]
            temp = nuclide.split('_')[2]
            is_water = (int(scaleid) == 1001) or (int(scaleid) == 1002)
            if (int(scaleid) < 100*1000) and not is_water:
                continue
            for mtlist in f[nuclide]['reaction_md']:
                mt = mtlist[0]
                if mt == 1007 or mt == 1008:
                    found_tsl = True
                    break
            if found_tsl:
                if last_scaleid == scaleid:
                    temp = str(float(temp))
                    tsl_table[last_scaleid_pos][2] += ", {}".format(temp)
                else:
                    name = stdcomp.name[stdcomp.scaleid==int(scaleid)].to_numpy()[0]
                    temp = str(float(temp))
                    tsl_table.append([str(int(scaleid)),name,temp,''])
                    last_scaleid_pos += 1
                last_scaleid = scaleid
    print("Number of TSLs found = ",last_scaleid_pos)
    tsl_df = pd.DataFrame(tsl_table)
    header = tsl_df.iloc[0]
    tsl_df = tsl_df.iloc[1:len(tsl_df)]
    tsl_df.columns = header
    tsl_df.to_markdown(output_file,tablefmt="grid",index=False,maxcolwidths=[None,None, 45,None])

def get_single_mat(root,endfmat):
    """
    Get the XML root object using python xml

    Parameters
    ----------
    root : Element
        The path to the XML file
    endfmat : str
        The ENDF MAT number identifying the element we want

    Returns
    -------
    child : Element
        A child Element object from the root Element
    """
    for child in root:
        endf = child.attrib['endf']
        if endf == endfmat:
            return child

def get_fastmat_thermal(root,realza,endfmat):
    """
    Read XML config file from SCALE data repo and get the SCALEID
    and ENDF MAT number for nuclei listed under "thermal" to find
    the MAT number for fast evaluation

    Parameters
    ----------
    root : Element object
        Element object from an ElementTree object from the python xml 
        package
    realza : str
        string of an int for the ZA of a material in ENDF format
    endfmat : str
        string of int for the ENDF MAT number

    Returns
    -------
    endf : str
        The MAT number for the fast evaluation
    """
    for child in root:
        # print(child.tag)
        if child.tag == 'thermal':
            for grandchild in child:
                # print(grandchild.keys())
                temp_realza = grandchild.get('realza')
                endf = grandchild.get('endf')
                # print(temp_realza,endf)
                if temp_realza == realza and endf == endfmat:
                    for greatgrandchild in grandchild:
                        endf = greatgrandchild.get('endf')
                        return endf

def get_xml_root(filename):
    """
    Get the XML root object using python xml

    Parameters
    ----------
    filename : str
        The path to the XML file

    Returns
    -------
    root : Element
        The root Element object from ElementTree
    """
    tree = ET.parse(filename)
    root = tree.getroot()
    return root

def get_scaleza_name_specialNuclei(root,realza,endfmat):
    """
    Read XML config file from SCALE data repo and get the SCALEID
    and ENDF MAT number for nuclei listed under "specialNuclei"

    Parameters
    ----------
    root : Element object
        Element object from an ElementTree object from the python xml 
        package
    realza : str
        string of an int for the ZA of a material in ENDF format
    endfmat : str
        string of int for the ENDF MAT number

    Returns
    -------
    scaleza, name : tuple
        A tuple of strings, the SCALEID and the name for SCALE

    Examples
    --------
    >>> import nuctools as nuc
    >>> e80root = nuc.get_xml_root('ENDF-8.0.xml_config')
    >>> nuc.get_scaleza_name_specialNuclei(e80root,'1001','125')
    """
    for child in root:
        #print(child.tag)
        if child.tag == 'specialNuclei':
            for grandchild in child:
                #print(grandchild.keys())
                temp_realza = grandchild.get('realza')
                endf = grandchild.get('endf')
                #print(temp_realza,endf)
                if temp_realza == realza and endf == endfmat:
                    scaleza = grandchild.get('scaleza')
                    name = grandchild.get('name')
                    return scaleza,name
def get_scaleza_name_thermal(root,realza,endfmat):
    """
    Read XML config file from SCALE data repo and get the SCALEID
    and ENDF MAT number for nuclei listed under "thermal"

    Parameters
    ----------
    root : Element object
        Element object from an ElementTree object from the python xml 
        package
    realza : str
        string of an int for the ZA of a material in ENDF format
    endfmat : str
        string of int for the ENDF MAT number

    Returns
    -------
    scaleza, name : tuple
        A tuple of strings, the SCALEID and the name for SCALE

    Examples
    --------
    >>> import nuctools as nuc
    >>> e80root = nuc.get_xml_root('ENDF-8.0.xml_config')
    >>> nuc.get_scaleza_name_thermal(e80root,'126','425')
    """
    names,scalezas = [],[]
    for child in root:
        # print(child.tag)
        if child.tag == 'thermal':
            for grandchild in child:
                # print(grandchild.keys())
                temp_realza = grandchild.get('realza')
                endf = grandchild.get('endf')
                # print(temp_realza,endf)
                if temp_realza == realza and endf == endfmat:
                    for greatgrandchild in grandchild:
                        scaleza = greatgrandchild.get('scaleza')
                        name = greatgrandchild.get('name')
                        scalezas.append(scaleza)
                        names.append(name)
                    return scalezas,names
def get_scaleza_name_metastable(root,realza,endfmat):
    """
    Read XML config file from SCALE data repo and get the SCALEID
    and ENDF MAT number for nuclei listed under "metastable"

    Parameters
    ----------
    root : Element object
        Element object from an ElementTree object from the python xml 
        package
    realza : str
        string of an int for the ZA of a material in ENDF format
    endfmat : str
        string of int for the ENDF MAT number

    Returns
    -------
    scaleza, name : tuple
        A tuple of strings, the SCALEID and the name for SCALE

    Examples
    --------
    >>> import nuctools as nuc
    >>> e80root = nuc.get_xml_root('ENDF-8.0.xml_config')
    >>> nuc.get_scaleza_name_metastable(e80root,'61148','6153')
    """
    for child in root:
        # print(child.tag)
        if child.tag == 'metastable':
            for grandchild in child:
                temp_realza = grandchild.get('realza')
                endf = grandchild.get('endf')
                # print(temp_realza,endf)
                if temp_realza == realza and endf == endfmat:
                    scaleza = grandchild.get('scaleza')
                    return scaleza
def append_xml_to_table(table,prot_dict,root,configroot,elibrary,additional_lib=False):
    """
    Append row of values of interest for table based on XML files

    Parameters
    ----------
    table : DataFrame
        DataFrame with members: "scaleid","name","gamprod","bondfac","gamxs","elibrary"
    prot_dict : dict
        python dictionary associating chemical symbols (e.g. Li) with number of protons
        (e.g. 3)
    root : Element
        root of main XML file listing for SCALE data, make with function `get_xml_root` above
    configroot : Element
        root of XML config file, make with function `get_xml_root` above
    elibrary : str
        Which endf library name you want to put in the table
    additional_lib : bool, optional
        is this the first lib appending to table
    
    Returns
    -------
    None : None
        It modifies the table in the input parameter

    """
    for child in root:
        scaleids,names=None,None # for thermal
        za = child.attrib['za']
        meta = child.attrib['metaStable']
        endf = child.attrib['endf']
        awi = child.attrib['awi']
        if za=='1':
            continue # skip neutron eval
        if awi=='0.0':
            continue # skip photo-atomic
        
        # gamma production
        try:
            awp0 = child.attrib['AWP0']
        except:
            awp0 = "no"
        # is TSL
        try:
            file7 = child.attrib['file7']
        except:
            file7 = "no"
        
        tag = child.attrib['tag']
        if file7 == "no":
            name = tag
            scaleid = za
        else:
            scaleids,names = get_scaleza_name_thermal(configroot,za,endf)

        # is metastable
        if meta == "true" and file7 == "no":
            scaleid = get_scaleza_name_metastable(configroot,za,endf)
        
        # special nuclei
        if za == '1001' and endf == '125':
            scaleid,name = get_scaleza_name_specialNuclei(configroot,za,endf)
        if za == '1002' and endf == '128':
            scaleid,name = get_scaleza_name_specialNuclei(configroot,za,endf)
    
        # ---- if TSL
        if file7 == "yes":
            for i in range(len(scaleids)):
                scaleid = scaleids[i]
                name = names[i]
                fastendf = get_fastmat_thermal(configroot,za,endf)
                fastelem = get_single_mat(root,fastendf)
                prot = int(fastelem.attrib['za'])//1000
                gamxs = prot_dict[str(prot)].lower()
                gamprod = "yes"
                bondfac = "yes"
                if additional_lib:
                    if scaleid in table['scaleid'].unique():
                        table.loc[table.scaleid==scaleid,'elibrary'] += ", " + elibrary
                    else:
                        # graphite's can still be called using same name
                        if '6000' in scaleid: 
                            elib = elibrary + "*"
                            print(elib)
                        else:
                            elib = elibrary
                        table.loc[len(table)] = [scaleid,name,gamprod,bondfac,gamxs,elib]
                else:        
                    table.loc[len(table)] = [scaleid,name,gamprod,bondfac,gamxs,elibrary]
        # ---- if *not* TSL
        else:
            prot = int(za)//1000
            gamxs = prot_dict[str(prot)].lower()
            gamprod = awp0
            bondfac = "yes"
            if additional_lib:
                if scaleid in table['scaleid'].unique():
                    table.loc[table.scaleid==scaleid,'elibrary'] += ", " + elibrary
                else:
                    # graphite's can still be called using same name
                    if '6000' in scaleid: 
                        elib = elibrary + "*"
                        print(elib)
                    else:
                        elib = elibrary
                    table.loc[len(table)] = [scaleid,name,gamprod,bondfac,gamxs,elib]
            else:        
                table.loc[len(table)] = [scaleid,name,gamprod,bondfac,gamxs,elibrary]










import pandas as pd
import numpy as np
import scipy.integrate as intgr
import time

# run all functions?
RUNALL = False

__all__ = ["c_edge","threshFinder",
           "decayFinder","six","four","two","sint",'ten','ten_exp',
           "noisy_gauss","egrp_avg","giveIt"]

mnc2 = 939565413.3  # NIST neutron mass [eV]
c = 299.792458      # NIST speed of light [m/us]

###### Functions ##############################################################

def c_edge(Eg):
    """
    Compute the Compton edge energy associatedwith a given 
    gamma energy
    
    Parameters
    ----------
    Eg : float
        Energy of gamma that compton scatters
    
    Returns
    -------
    Ec : float
        Energy of Compton edge

    Notes
    -----
    .. math:: E_c(E_\\gamma) = \\frac{2.0 E_\\gamma^2}{511.0+2.0E_\\gamma}
    """
    Ec = 2.0*Eg**2/(511.0+2.0*Eg)
    return Ec

#------------------------------------------------------------------------------
# input: pulse vector (floats), threshold where pulse
#        increases above 20% for example
# output: index where threshold is FIRST crossed
def threshFinder(pulse,threshold):
    """
    For a positive "pulse", define a threshold and return the 
    index immediately after that threshold is passed.

    Parameters
    ----------
    pulse: array_like
        1-d set of values that make up the ordinates of a 
        positive pulse
    threshold : float
        A value > the minimum pulse value, and < the max pulse
        value
    
    Returns
    -------
    index : int
        The index of the pulse value immediately after the 
        pulse value is greater than the threshold

    """
    index = 0
    for i in range(len(pulse)):
        if pulse[i] > threshold:
            index = i
            break
    return index
#------------------------------------------------------------------------------
# input: pulse vector (floats), threshold where pulse
#        decreases below 20% for example
# output: index where threshold is SECOND crossed
def decayFinder(pulse,threshold):
    """
    For a positive "pulse", define a threshold and return the 
    index immediately after that threshold is passed for a 
    second time. 

    Parameters
    ----------
    pulse: array_like
        1-d set of values that make up the ordinates of a 
        positive pulse
    threshold : float
        A value > the minimum pulse value, and < the max pulse
        value
    
    Returns
    -------
    index : int
        The index of the pulse value immediately after the 
        pulse value is < than the threshold, if the 
        threshold has already been passed once.
    """
    index = 0
    flag = False
    peakIndex = max(range(len(pulse)), key=pulse.__getitem__)
    for i in range(len(pulse)):
        if (flag == True) and (pulse[i] < threshold) and (i > peakIndex):
            index = i
            break
        if pulse[i] > threshold:  # flag the first crossing
            flag = True
    return index
#------------------------------------------------------------------------------
# string formatter
def six(x):                     # (6 digits . 6 digits) float
    return "{:6.6f}".format(x)
def four(x):                     # (6 digits . 4 digits) float
    return "{:6.4f}".format(x)
def two(x):                     # (6 digits . 2 digits) float
    return "{:6.2f}".format(x)
def sint(x):                    # int
    return "{:d}".format(x)
def ten(x):
    return "{:9.5f}".format(x) # (10 digits total, 4 digits after dec.)
def ten_exp(x):
    return "{:9.3e}".format(x) # (10 digits total, 4 digits after dec.)
#------------------------------------------------------------------------------

def noisy_gauss(x,mu,sig):
    """
    Generate a Gaussian distribution with noise.

    Gaussian is generated with user input average and 
    standard deviation, noise is added to all data pts.
    with the numpy random package.

    Parameters
    ----------
    x : numpy array 
        Vector of absiccae to plot the gaussian along
    mu : float
        The average value of the gauss, where the peak
        will occur.
    sig : float
        The standard deviation from the average, varying
        the width of the "bell" curve.
    
    Returns
    -------
    gauss : numpy array
        The ordinates of same length x, which give the
        Gaussian or normal distribution for the specified
        parameters.
    
    Examples
    --------
    >>> x = np.linspace(0,100,100)
    >>> mu,sig = 50,10
    >>> gauss = noisy_gauss(x,mu,sig)
    >>> plt.figure(3)
    >>> plt.plot(x,gauss,color="k",ls=" ",marker=".")
    >>> plt.show()

    """
    # gauss curve
    gauss = (1/(2*np.pi*sig**2)**0.5*np.exp(-(x-mu)**2/(2*sig**2))
             +np.random.rand(len(x))*0.01)
    
    return gauss
#------------------------------------------------------------------------------
def egrp_avg(E,Eg,data,err_data):
    """
    Group quantities and average them along specified
    energy grid Eg.

    Grouping is done with integration by scipy quad()
    function and division by group size.

    Parameters
    ----------

    E : array_like
        Energy grid data is currently on
    Eg : array_like
        Grouped energy grid
    data : array_like
        Quantity to be group averaged
    err_data : array_like 
        The error on the data array. Must be same
        length as data array.

    Returns
    -------

    gdata : pandas DataFrame
        The grouped data is attribute 'group_avg', and the
         **estimated** absolute error on the grouped data
         is attribute 'err_group_avg'.

    Notes
    -----
    The estimated error on the grouped data is simply the error
    on a simple average on the group. The actual grouping is done
    by the scipy.integrate.quad function.

    .. math:: \\frac{\\Delta \\langle x\\rangle}{\\langle x\\rangle}= \\frac{\\sqrt{\\sum^N (\\Delta x )^2}}{\\sum^Nx}  

    """
    # cast inputs as numpy arrays
    E,Eg,data = np.array(E),np.array(Eg),np.array(data)
    # if E is not in ascending order throw error
    if any(np.diff(E) < 0) or any(np.diff(Eg) < 0):
        raise ValueError("E,Eg must be in ascending order.")
    # create grid with 20 times the energy points (linear)
    interpE = np.linspace(0,max(E),len(E)*20)
    # create function to integrate over
    integ = lambda interpE: np.interp(interpE,E,data)
    # make energy group average array 
    Eg_avg = np.zeros(len(Eg))
    rel_err = np.zeros(len(Eg))
    # ---------------------------------------
    # --- error estimation ------------------
    # ---------------------------------------
    df = pd.DataFrame({'E':E,'data':data,'err_data':err_data})
    # for each energy group, integrate and normalize
    for i in range(len(Eg)):
        Eg_avg[i] = intgr.quad(integ,Eg[i-1],Eg[i])[0]/(Eg[i]-Eg[i-1])
        # --- error ----------
        squared_error = df.err_data[(df.E>Eg[i-1]) & (df.E<Eg[i])]**2
        simple_grp = np.sum(df.data[(df.E>Eg[i-1]) & (df.E<Eg[i])])
        rel_err[i] = np.sqrt(squared_error.sum())/simple_grp


    return pd.DataFrame({'group_avg':Eg_avg,'err_group_avg':rel_err*Eg_avg})

#------------------------------------------------------------------------------
def frohner_cor(sig1,sig2,n1,n2):
    """
    Takes cross-sections [barns] and atom densities [atoms/barn] for 
    two thicknesses of the same sample, and returns extrapolated cross
    section according to Frohner.

    Parameters
    ----------
    sig1 : array_like
        Cross section of the thinner of the two samples.
    sig2 : array_like
        Cross section of the thicker of the two samples.
    n1 : float
        Atom density of the thinner sample
    n2 : float 
        Atom density of the thicker sample

    Returns
    -------
    sig0 : array_like
        The extrapolated cross section from sig1 and sig2
    """

    return (n2*sig1-n1*sig2)/(n2-n1)

#------------------------------------------------------------------------------
def frohner_cor_3rd_order(sig1,sig2,sig3,n1,n2,n3):
    """
    Takes cross-sections [barns] and atom densities [atoms/barn] for 
    three thicknesses of the same sample, and returns extrapolated 
    cross section according to Frohner.

    Parameters
    ----------
    sig1 : array_like
        Cross section of the thinnest of the three samples.
    sig2 : array_like
        Cross section of the mid-thickness of the three samples.
    sig3 : array_like
        Cross section of the thickest of the three samples.
    n1 : float
        Atom density of the thinnest sample
    n2 : float 
        Atom density of the mid-thickness sample
    n3 : float
        Atom density of the thickest sample

    Returns
    -------
    sig0 : array_like
        The extrapolated cross section from sig1, sig2, and sig3
    """

    # two terms in the numerator
    numer1 = (n1*sig2-n2*sig1)*(n3**2-n1**2-(n1-n3)/(n1-n2)*(n2**2-n1**2))
    numer2 = (n1*n2**2-n1**2*n2)*(sig3-sig2-(n1-n3)/(n1-n2)*(sig2-sig1))
    denom = (n1-n2)*(n3**2-n1**2) - (n1-n3)*(n2**2-n1**2)
    
    return (numer1-numer2)/denom

#############################################################

#############################################################

def giveIt():
    return 900



import pandas as pd

__all__ = ["read_depletion_history"]

def read_depletion_history(triton_inp):
    """
    Read the depletion history from a SCALE/TRITON input file
    and put it into a pandas DataFrame

    Parameters
    ----------
    triton_inp : str
        The path and filename for the triton input file

    Returns
    -------
    history : DataFrame
        A DataFrame with columns: 'power','burn','str','down','nlib','end'
        where `power` is the power in MW and `burn` is the days burning
        the fuel, and `down` is the downtime between burns
    """
    skip_lines = 0
    num_lines = 0
    with open(triton_inp,'r') as f:
        lines = f.readlines()
    for i,line in enumerate(lines):
        if "read burndata" in line:
            skip_lines = i+1
        if "end burndata" in line and skip_lines != 0:
            num_lines = i-skip_lines
    if skip_lines == 0 and num_lines == 0:
        raise ValueError("Could not find burndata in input file")

    history = pd.read_csv(triton_inp,skiprows=skip_lines,nrows=num_lines,delim_whitespace=True,
                          names=['power','burn','str','down','nlib','end'])
    history['power'] = history.power.str.replace('power=','').astype(float)
    history['burn'] = history.burn.str.replace('burn=','').astype(float)
    print("Total burnup = {:.4f} GWd/t".format((history.power*history.burn).sum()/1000))

    return history




import matplotlib.pyplot as plt
import mpl_toolkits.axisartist as aa

__all__ = ['set_ticks_1axes','set_ticks_yt']

def set_ticks_1axes(axes_name):
	"""
	Used to set the ticks in the style Yaron Danon requires
	for his graduate students.
	
	Parameters
	----------
	axes_name : matplotlib axes object
	    This should be an axisartist axes object created like
	    in the example shown below
	
	Returns
	-------
	nothing
	
	Examples
	--------
	>>> import matplotlib.pyplot as plt
	>>> import mpl_toolkits.axisartist as aa
	>>> fig = figure(1)
	>>> ax1 = aa.Subplot(fig,111)
	>>> fig.add_subplot(ax1)
	>>> set_ticks_1axes(ax1)
	
	"""
	axes_name.axis["bottom","left"].major_ticks.set_tick_out(True)
	axes_name.axis["top","right"].major_ticks.set_tick_out(False)
	axes_name.axis["bottom","left"].minor_ticks.set_tick_out(True)
	axes_name.axis["top","right"].minor_ticks.set_tick_out(False)


def set_ticks_yt(axes_name1,axes_name2):
	"""
	Used to set the ticks in the style Yaron Danon requires
	for his graduate students. This in particular is designed
	for plotting yield and transmission on top of each other.
	
	Parameters
	----------
	axes_name1 : matplotlib axes object
	    This should be an axisartist axes object created like
	    in the example shown below
	axes_name2 : matplotlib axes object
	    This should be an axisartist axes object created like
	    in the example shown below
	
	axes_name : matplotlib axes object
	    This should be an axisartist axes object created like
	    in the example shown below
	
	Returns
	-------
	nothing
	
	Examples
	--------
	>>> import matplotlib.pyplot as plt
	>>> import mpl_toolkits.axisartist as aa
	>>> fig = figure(1)
	>>> ax1 = aa.Subplot(fig,211)
	>>> fig.add_subplot(ax1),fig.add_subplot(ax2)
	>>> set_ticks_yt(ax1,ax2)
	
	"""
	axes_name1.axis["bottom"].major_ticks.set_tick_out(False)
	axes_name1.axis["left"].major_ticks.set_tick_out(True)
	axes_name1.axis["bottom"].minor_ticks.set_tick_out(False)
	axes_name1.axis["left"].minor_ticks.set_tick_out(True)
	axes_name1.axis["top"].major_ticks.set_tick_out(False)
	axes_name1.axis["right"].major_ticks.set_tick_out(False)
	axes_name1.axis["top"].minor_ticks.set_tick_out(False)
	axes_name1.axis["bottom"].major_ticklabels.set_visible(False)
	#-----------------------------
	axes_name2.axis["bottom"].major_ticks.set_tick_out(True)
	axes_name2.axis["left"].major_ticks.set_tick_out(True)
	axes_name2.axis["bottom"].minor_ticks.set_tick_out(True)
	axes_name2.axis["left"].minor_ticks.set_tick_out(True)
	axes_name2.axis["top","right"].major_ticks.set_tick_out(False)
	axes_name2.axis["top","right"].minor_ticks.set_tick_out(False)
	axes_name2.axis["bottom"].major_ticklabels.set_visible(True)


import os,sys,time
import pandas as pd
import numpy as np
import subprocess as sub
import yaml
from pathlib import Path
from . import funky as fk 
from . import sam_tools as st

__all__ = ["sesh_fitacs","write_pars_fitacs","apply_sesh_corr",
           "write_sesh_pars","read_fitacs_par","create_new_sesh_mcfile"]

def sesh_fitacs(inp_file_name,data):

    """
    This function performs the iteration loop for SESH and FITACS

    SESH and FITACS are used together to correct experimental 
    cross section data and fit that data, respectively. Both 
    programs require average resonance parameters and to use them
    in concert requires that the input to both be nearly the 
    same. Therefore an iteration must be performed feeding the 
    input and output between the SESH and FITACS programs. The
    sesh_fitacs() function performs this.

    Parameters
    ----------
    inp_file_name : str
        The input file must be a yaml format file (.yml or .yaml), with 
        variables defined as in the example file.
    data : list
        A list of pandas DataFrames, which contain very specific column
        names. This list of DataFrames contain the experimental data 
        used to fit in FITACS. The order of this list must be the same
        order as the lists given in the input file inp_file_name.

    Returns
    -------
    none : ascii files
        This function will generate a lot of ascii files output from 
        SESH and FITACS. These will have naming conventions that are
        used in the function. The final fit with FITACS is after 
        the convergence criterion has been met and completes the 
        while loop.

    Examples
    --------
    >>> xs_col = ['e','cs','dcs']
    >>> tr_col = ['e','t','dt']
    >>> totxs_ta1 = pd.read_csv("ta1_sig.dat"  ,names=xs_col,delim_whitespace=True)
    >>> trans_ta1 = pd.read_csv("ta1_trans.dat",names=tr_col,delim_whitespace=True)
    >>> data1 = pd.concat([totxs_ta1,trans_ta1[['t','dt']]],axis=1)
    >>>
    >>> totxs_ta3 = pd.read_csv("ta3_sig.dat"  ,names=xs_col,delim_whitespace=True)
    >>> trans_ta3 = pd.read_csv("ta3_trans.dat",names=tr_col,delim_whitespace=True)
    >>> data3 = pd.concat([totxs_ta3,trans_ta3[['t','dt']]],axis=1)
    >>>
    >>> totxs_ta6 = pd.read_csv("ta6_sig.dat"  ,names=xs_col,delim_whitespace=True)
    >>> trans_ta6 = pd.read_csv("ta6_trans.dat",names=tr_col,delim_whitespace=True)
    >>> data6 = pd.concat([totxs_ta6,trans_ta6[['t','dt']]],axis=1)
    >>>
    >>> capxs_ta1 = pd.read_csv("capxs_ta1.dat",names=xs_col,delim_whitespace=True)
    >>> capxs_ta2 = pd.read_csv("capxs_ta2.dat",names=xs_col,delim_whitespace=True)
    >>>
    >>> data = [data1,data3,data6,capxs_ta1,capxs_ta2]
    >>> 
    >>> directory = '/Users/jesse/Dropbox/ta_urr_fitting/'
    >>> inp_file = directory+'sesh_fitacs_inp.yml'
    >>>
    >>> nuc.sesh_fitacs(inp_file,data)

    Notes
    -----
    This is to be more fully documented in the greater nuctools documentation.

    """
    
    start_tot = time.time()
    
    # ------------------------------------------------------------------------
    # import all the variables from the input file into a dict
    # ------------------------------------------------------------------------
    with open(inp_file_name,'r+') as f:
        inp = yaml.load(f)
    
    wdir          = inp['workdir']
    fitacs_indir  = wdir+inp['fitacs_indir']
    fitacs_outdir = wdir+inp['fitacs_outdir']
    sesh_indir    = wdir+inp['sesh_indir']
    sesh_outdir   = wdir+inp['sesh_outdir']
    ebd           = inp['e_bounds']
    thickness     = inp['samp_thick']
    cap_bool      = inp['cap_bool']
    data_names    = inp['data_names']
    data_dir      = wdir+inp['data_dir']
    # ------------------------------------------------------------------------
    
    # ------------------------------------------------------------------------
    # iterate until fitted FITACS pars are within the bounds of experimental 
    # uncertainty of the SESH pars used to correct the data input to FITACS
    # ------------------------------------------------------------------------
    #
    # ----------------------
    # set par filename as the starting FITACS par file
    # ----------------------
    temp_filename = fitacs_indir+inp['fitacs_par_name']
    
    # ----------------------
    # Calculate correction factor for first set of pars?
    # Often the beginning set has been calculated by a previous run
    # ----------------------
    calc_first_corr = inp['calc_first_corr']
    
    
    iteration = 0
    sesh_fitacs_pars_not_within_unc = True
    
    while sesh_fitacs_pars_not_within_unc:
        start_it = time.time()
        
        pars = read_fitacs_par(temp_filename,num_e_regions=3)
    
        print('#######################################')
        print('Iteration ',iteration)
        print('#######################################')
        
        # convergence boolean is tested in loop
        if iteration > 0:
            sesh_fitacs_pars_not_within_unc = False

        for e in range(inp['fitacs_numE_regions']):
            start_e = time.time()
            print('----------------')
            print('Energy region',e)
            print('----------------')
            # ----------------------
            # create a sesh file for each E_region from the pars read from fitacs
            # ----------------------
            temp_sesh_filename = sesh_indir+inp['sesh_ifile_name']
            write_sesh_pars(temp_sesh_filename,
                            temp_sesh_filename+'_e{}_it{}'.format(e,iteration),
                            pars[e*3:e*3+3].reset_index(),inp['Rp'])

            # ----------------------
            # write a new interactive input file for each e region
            # ----------------------
            
            temp_int_input = sesh_indir+inp['sesh_int_input']
            with open(temp_int_input,'r+') as f:
                with open(temp_int_input+'{}'.format(e),'w+') as f2:
                    for line in f:
                        # remove quotes, trailing whitespace, and newlines
                        line = line.replace('\"','').strip()
                        f2.write('\"'+line+'_e{}_it{}\"\n'.format(e,iteration))

            # ----------------------
            # run the SESH program
            # ----------------------
            if iteration == 0 and calc_first_corr == False:
                calculate = False
                print("Skipping first SESH calc, as it already exists.")
                print("This can be controlled by input.")
            else:
                calculate = True
            if calculate:
                # -> create a process
                run_sesh = sub.Popen('sesh',stdin=sub.PIPE,stdout=sub.PIPE,stderr=sub.PIPE)
                # -> make a byte string to send to process
                with open(temp_int_input+'{}'.format(e),'r+') as f:
                    byte_str_inp = str.encode(f.read())
                # -> send the byte string to SESH and run
                print("SESH MC calculation running...",time.ctime())
                out, err = run_sesh.communicate(input=byte_str_inp)
                #print(err)
                if (Path(wdir+'fort.99').exists()):
                    Path(wdir+'fort.99').unlink()
                print("SESH run completed.")

            # ----------------------
            # correction file must be interpreted from the results in output
            # ----------------------
            temp_corr_file = sesh_outdir+inp['sesh_cor']+'_e{}_it{}'.format(e,iteration)
            temp_output = sesh_outdir+inp['sesh_output']+'_e{}_it{}'.format(e,iteration)
            create_new_sesh_mcfile(temp_output,temp_corr_file)
            
            # ----------------------
            # Calculate difference between current and previous correction factors
            # ----------------------
            if iteration > 0:
                prev_corr_file = sesh_outdir+inp['sesh_cor']+'_e{}_it{}'.format(e,iteration-1)
                colnames = ['e','avtot','avcap','sscf','davtot','davcap','dsscf','thick','samp_type']
                current = pd.read_csv(temp_corr_file,delim_whitespace=True,
                                      names=colnames,header=0)
                previous = pd.read_csv(prev_corr_file,delim_whitespace=True,
                                       names=colnames,header=0)
                
                mean_change = np.mean(abs(current.sscf-previous.sscf)/previous.sscf)
                print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")
                print(fk.four(mean_change*100),"% change from previous correction factor")
                print("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<")
                
                if mean_change > 0.01:
                    sesh_fitacs_pars_not_within_unc = True

            # ----------------------
            # apply correction to specific energy region of experimental data for 
            # each sample of thickness i,thick
            # ----------------------
            for i,thick in enumerate(thickness):
                # if capture sample
                if cap_bool[i]:
                    new_xs,new_dxs = apply_sesh_corr(temp_corr_file,
                                                     [data[i].e,data[i].cs,data[i].dcs],
                                                     thick,capture=True)
                # if transmission sample
                else:
                    new_xs,new_dxs = apply_sesh_corr(temp_corr_file,
                                                     [data[i].e,data[i].t,data[i].dt],
                                                     thick,capture=False)

                corr = pd.DataFrame({
                    'e': data[i].e,
                    'cs': new_xs,
                    'dcs': new_dxs,
                })
                #print(corr.cs)
                corr = corr[(corr.e > ebd[e]) & (corr.e < ebd[e+1])]
                # set data[thickness i] to corrected data
                data[i].cs[(data[i].e > ebd[e]) & (data[i].e < ebd[e+1])] = corr.cs
                data[i].dcs[(data[i].e > ebd[e]) & (data[i].e < ebd[e+1])] = corr.dcs

            finish_e = time.time()
            print("Time for E region {} = {} sec".format(e,fk.two(finish_e-start_e)))

        
        
        if sesh_fitacs_pars_not_within_unc is False:
            print("\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\")
            print("/////////////////////////////////////////////////////////////")
            print("SESH correction factor changed {:10.4f}% between iterations.".format(mean_change*100))
            print("Convergence achieved.")
            print("Final fit with FITACS...")
            print("\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\")
            print("/////////////////////////////////////////////////////////////")
        
        # ----------------------
        # Create new data files to fit w/ FITACS
        # ----------------------
        for i,thick in enumerate(thickness):
            filename = data_dir+data_names[i]
            st.fmt_ten([data[i].e,data[i].cs,data[i].dcs],
                        filename,tot= not cap_bool[i])

        # ----------------------
        # write a new interactive input file for FITACS
        # ----------------------
        temp_int_input = fitacs_indir+inp['fitacs_int_input']
        with open(temp_int_input,'r+') as f:
            with open(temp_int_input+'{}'.format('_'),'w+') as f2:
                for i,line in enumerate(f):
                    filename = data_dir+data_names[i-2]
                    if i < 2:
                        f2.write(line)
                    else:
                        f2.write(filename+'\n')


        # ----------------------
        # Fit corrected data with FITACS
        # ----------------------
        # -> create a process
        run_fitacs = sub.Popen('sammy',stdin=sub.PIPE,stdout=sub.PIPE,stderr=sub.PIPE)
        # -> make a byte string to send to process
        with open(temp_int_input+'{}'.format('_'),'r+') as f:
            byte_str_inp = str.encode(f.read())
        # -> send the byte string to SESH and run
        out, err = run_fitacs.communicate(input=byte_str_inp)
        #print(err)
        print('\n+++++++++++++++++++++')
        print("FITACS run completed.")
        print('+++++++++++++++++++++')

        # ----------------------
        # Rename FITACS output and delete unnecessary files
        # ----------------------
        fit_out = inp['fitacs_outdir']
        p = Path(wdir)
        for item in list(p.glob('*.lst')):
            item.rename(Path(item.parent, "{}{}".format(fit_out,
                                                        item.stem+item.suffix)))
        # .unlink() deletes the files 
        for item in list(p.glob('*.odf')):
            item.unlink()
        for item in list(p.glob('*.plt')):
            item.unlink()
        for item in list(p.glob('*.DAT')):
            item.unlink()
        Path(wdir+'SAMMY.IO').unlink()

        # rename files
        par_path = Path(wdir+'SAMMY.PAR')
        lpt_path = Path(wdir+'SAMMY.LPT')
        cov_path = Path(wdir+'SAMMY.COV')
        par_path_targ = Path(wdir+fit_out+'fit.par')
        lpt_path_targ = Path(wdir+fit_out+'fit.lpt')
        cov_path_targ = Path(wdir+fit_out+'fit.cov')
        par_path.rename(par_path_targ)
        lpt_path.rename(lpt_path_targ)
        cov_path.rename(cov_path_targ)
        
        # failsafe: only allow 4 iterations
        if iteration > 3:
            print("\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\")
            print("///////////////////////////////////////")
            print("Max iterations reached, no convergence.")
            print("\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\")
            print("///////////////////////////////////////")
            break
        
        # ----------------------
        # If another iteration is necessary, create a new par file
        # ----------------------
        if sesh_fitacs_pars_not_within_unc:
            # ----------------------
            # read the fitted parameters 
            # ----------------------
            pars_new = read_fitacs_par(wdir+fit_out+'fit.par',num_e_regions=3)
            
            p = Path(temp_filename)
            # create new FITACS input par file with new pars
            new_filename = (str(p.parent)+"/"+str(p.stem)+
                            "_it{}".format(iteration)+str(p.suffix))
            
            write_pars_fitacs(temp_filename,new_filename,pars_new)
            
            temp_filename = new_filename
            
        finish_it = time.time()
        it_time = fk.two(finish_it-start_it)
        print("Time for iteration {} = {}".format(iteration,it_time))
        iteration += 1
            
    
    finish_tot = time.time()
    tot_time = finish_tot-start_tot
    print("Total time = {}".format(tot_time))
    
    with open('run_{}_stats.txt'.format(time.ctime()),'w+') as f:
        f.write('Various run statistics:\n')
        f.write('-----------------------\n')
        f.write("Total time = {}\n".format(tot_time))
        f.write('Total number of iterations = {}\n'.format(iteration))

def write_pars_fitacs(init_filename,new_filename,pars):
    """
    Write average parameters from memory into a pre-formatted 
    FITACS par file.
    
    Parameters
    ----------
    init_filename : str
        The initial par file name (full or relative path). This
        should have the exact same number of parameters for each
        partial wave as the pars DataFrame input
    new_filename : str
        The name of the file to be written to (full or relative 
        path). This file will be the same as init_filename but 
        with different parameters
    pars : pandas DataFrame
        The pars DataFrame has average resonance parameters 
        necessary to run the SESH and FITACS codes. This variable
        has columns of data described in read_fitacs_par()
    
    Returns
    -------
    none : ascii file
        Replaces the parameters in init_filename and re-writes the
        file to new_filename.
    
    """
    
    def ten(number):
        return '{:10.6f}'.format(number)
    
    with open(init_filename,'r+') as f: file = f.readlines()
    with open(new_filename,'w+') as f:
        k = -1
        # skip this many lines before writing to new file
        write_line = 0
        for line in file:
            # found the average resonance parameters
            if ('stre' in line) or ('STRE' in line):
                f.write(line)
                k+=1
                # for all ang mom L
                for j in range(int(max(pars.L)+1)):
                    f.write((ten(pars.stren[k+j])+ten(pars.dstren[k+j])+
                      ten(pars.dist_R[k+j])+ten(pars.ddist_R[k+j])+
                      ten(pars.gam_g[k+j])+ten(pars.dgam_g[k+j])))
                    if j==0:
                        f.write(ten(pars.D[0])+'\n')
                    else:
                        f.write('\n')
                write_line = int(max(pars.L)+2)
            if write_line > 0:
                write_line-=1
                continue
            else:
                f.write(line)



def apply_sesh_corr(cor_file,data,N,capture=True):
    """
    Correct experimental data with a Monte Carlo correction from SESH

    Parameters
    ----------
    cor_file : str
        The full file path or relative to the correction factor file
    data : array-like
        2-d array with data[0] as energy; For capture, data[1] as cross section,
        data[2] as the absolute uncertainty in the cross section. If correcting 
        total cross section data[1] should be the observable (transmission), 
        data[2] is the absolute uncert. on the observable. Units for xs are
        barns.
    N : float
        The areal number density [at/barn] of the sample is needed if correcting
        the total cross section, but not for the capture cross section
    capture : bool, optional
        Whether the correction is being applied to capture or total cross
        section. If total, N must be specified

    Returns
    -------
    new_xs,new_dxs : tuple
        A tuple of numpy arrays containing the corrected cross section [barns]
        and the uncertainty on the corrected cross section [barns].

    Examples
    --------
    >>>
    
    Notes
    -----
    Capture

    .. math:: \\alpha = correction
    
    .. math:: \\sigma_{\\gamma,\\alpha} = \\sigma_{\\gamma}/\\alpha
    
    .. math:: \\Delta\\sigma_{\\gamma,\\alpha} = \\sigma_{\\gamma,\\alpha}\\sqrt{\\frac{\\Delta\\sigma}{\\sigma}^2+\\frac{\\Delta \\alpha}{\\alpha}^2}
    
    Trans
    
    .. math:: \\sigma_{t,\\alpha} = \\frac{-1}{N}ln\\left(\\frac{T}{\\alpha}\\right)
    
    .. math:: \\Delta\\sigma_{t,\\alpha} = \\frac{1}{N}\\sqrt{\\frac{\\Delta T}{T}^2+\\frac{\\Delta\\alpha}{\\alpha}^2}
    
    
    """
    # the column names 
    colnames = ['e','totxs','dtotxs','capxs','dcapxs','sscf',
                'dsscf','trans','dtrans','thick','samp_type']

    if (N is None) and (capture is False):
        raise ValueError("If correcting total xs, N must be specified.")
    
    # cast data to numpy 
    data = np.array(data)
    # read in the correction factor file
    sesh = pd.read_csv(cor_file,delim_whitespace=True,names=colnames,
                       header=0,float_precision='round_trip') 
    sesh = sesh[sesh.thick == N]
    
    # interpolate the correction onto the data energy grid
    cor = np.interp(data[0],sesh.e*1000,sesh.sscf)
    dcor = np.interp(data[0],sesh.e*1000,sesh.dsscf)/100 # percent
    
    # apply the correction
    if capture:
        new_xs = data[1]/cor
        new_dxs = new_xs*np.sqrt((data[2]/data[1])**2+(dcor)**2)
    else:
        new_xs = -1/N*np.log(data[1]/cor)
        new_dxs = 1/N*np.sqrt((data[2]/data[1])**2 + (dcor)**2)
        
    return new_xs,new_dxs

def write_sesh_pars(ifilename,ofilename,pars,Rp,max_orb_ang_mom=2):
    """
    Re-write the SESH input file with new resonance parameters.

    Parameters
    ----------
    ifilename : str
        The base file you want to re-write. Full path or relative.
    ofilename : str
        The new input file with updated parameters. Full path or relative
    pars : Pandas DataFrame
        DataFrame that is returned from the function read_fitacs_par() which
        is described below. DataFrame attributes (columns) must have the same
        names as described in read_fitacs_par()
    Rp : float
        The effective radius which will be used for the SESH calculation
    max_orb_ang_mom : int, optional
        The maximum orbital quantum angular momentum. SESH accepts avg 
        parameter values for each L value

    Returns
    -------
    None : ascii file
        Returns a re-written SESH ascii input file with new average parameters

    Examples
    --------

    >>> import nuctools as nuc
    >>> directory = '/Users/jesse/Dropbox/ta_urr_fitting/'
    >>> pars = nuc.read_fitacs_par(directory+'multi_fit.par',num_e_regions=3)
    >>> nuc.write_sesh_pars(directory+'input.data',directory+'input.data1',pars[0:3],7.8)

    """
    with open(ifilename,'r+') as f:
        # create a list of line strings
        file = f.readlines()
    with open(ofilename,'w+') as f:
        i = 0
        for line in file:
            # write the pars in file lines 3->(3+max orb ang momentum)
            if (i > 1) and (i <= 2+max_orb_ang_mom):
                # write the avg pars for each partial wave
                f.write(fk.ten(pars.gam_g[i-2])+' '
                        +fk.ten(pars.D[i-2])+' '
                        +fk.ten_exp(pars.stren[i-2])+' '
                        +fk.ten(0.0)+fk.ten(Rp)+fk.ten(1.0)+'\n')
                i += 1
                continue
            # re-write the lines for everywhere else
            f.write(line)
            i += 1


def read_fitacs_par(filename,num_e_regions=1,max_orb_ang_mom=2):
    """
    Function to read the average parameters from a FITACS/SAMMY par file

    The output par file from FITACS has average resonance parameters for
    multiple angular momentums, and multiple energy regions (if user 
    specified). This function reads the file based on the user input 
    for those two values.

    Parameters
    ----------
    filename : str
        The name of the file (full path or relative) to be read
    num_e_regions : int, optional
        The number of energy regions over which SAMMY/FITACS is fitting
        the average resonance parameters
    max_orb_ang_mom : int, optional
        The maximum orbital angular momentum that was considered when 
        running the SAMMY/FITACS program. A typical value is L=2.

    Returns
    -------
    pars : Pandas DataFrame
        A dataframe that contains the following attributes: 

        - L : the orbital angular momentum
        - D : the average level spacing
        - eregion : the number of energy regions fit
        - stren : the neutron strength function
        - dstren : the unc. on the neutron strenth function
        - dist_R : the distant level parameter R-inf
        - ddist_R : the unc. on dist_R
        - gam_g : the average radiation width 
        - dgam_g : the unc. on gam_g

    Examples
    --------

    >>> import nuctools as nuc
    >>> pars = nuc.read_fitacs_par('multi_fit.par',num_e_regions=3)
    >>> print(pars)

    """
    with open(filename,'r+') as f:
        res_pars_next = False
        pars_break = False
        df_len = num_e_regions*(max_orb_ang_mom+1)
        pars = pd.DataFrame({
            'L'      : np.zeros(df_len),
            'D'      : np.zeros(df_len),
            'eregion': np.zeros(df_len),
            'stren'  : np.zeros(df_len),
            'dstren' : np.zeros(df_len),
            'dist_R' : np.zeros(df_len),
            'ddist_R': np.zeros(df_len),
            'gam_g'  : np.zeros(df_len),
            'dgam_g' : np.zeros(df_len),
        })
        region = 0
        L = 0
        
        for i,line in enumerate(f):
            #print(i)
            if len(line) < 4:
                continue
                
            # remove the line endings and lines with only \n
            line = line.strip('\n')
            
            # test if resonance parameters are coming next
            if ('STRE' in line) or ('Stre' in line):
                #print("region = ",region,"\n---------")
                res_pars_next = True
                pars_break = False
                L = 0
                continue
            # test if we've reached the end of the res. pars 
            if '---' in line:
                pars_break = True
            if res_pars_next and not pars_break:
                # grab the level spacing for s-waves
                if L == 0:
                    D = float(line[60:70])
                # choose element to fill
                pos = region*3 + L
                # set the values
                pars.L[pos]       = L
                pars.D[pos]       = D
                pars.eregion[pos] = region
                pars.stren[pos]   = float(line[0:10])
                pars.dstren[pos]  = float(line[10:20])
                pars.dist_R[pos]  = float(line[20:30])
                pars.ddist_R[pos] = float(line[30:40])
                pars.gam_g[pos]   = float(line[40:50])
                pars.dgam_g[pos]  = float(line[50:60])
                
                # increment the orbital ang momentum value
                L += 1
            if res_pars_next and pars_break:
                res_pars_next = False
                # increment the energy region
                region += 1
                
    return pars


def create_new_sesh_mcfile(output_filename,clean_mc_filename,runtype='trans'):
    """
    Read the sesh output file defined by output_filename. Collect
    Monte Carlo run data including average capture cross section,
    average total cross section, self-shielding correction factor,
    and the error on all of these.

    Parameters
    ----------
    output_filename : str
        the name of the output file from a sesh run. Usually is
        output.data, but I suggest renaming it as sesh will 
        keep overwriting this file each time you use sesh.
    clean_mc_filename : str
        The name of the file you wish to write the MC info to.    
    runtype : str, optional
        The type of SESH run that was performed. Right now the 
        supported types are transmission (input must include 
        either "trans" or "TRANS" in string) and capture 
        (runtype is left at default None). If both transmission
        and capture are input to the SESH program, this function
        will decide where there is capture or transmission data
        being given in the output file and this input is 
        unnecessary.

    Returns
    -------
    None : ascii file
        Returns an ascii file with a descriptive header for capture.
    """
    # Choose the type of self-shielding that was calculated in SESH
    captype = True
    if 'trans' in runtype or "TRANS" in runtype:
        captype = False

    # open new output sesh file and collect data
    with open(clean_mc_filename,"w+") as fnew:
        # write a header
        fnew.write("Energy[keV] Av.Tot.xs[b] dAv.Tot.xs[b] Av.Cap.xs[b] dAv.Cap.xs[b]")
        fnew.write(" SSCor.Fac.[] dSSCor.Fac.[%] trans[] dtrans[] ")
        fnew.write("thick[at/b] samp_type\n")
        # open the file from sesh
        with open(output_filename,"r+") as f:
            # create array of line strings from sesh file
            f_string_array = f.readlines()
            start_of_mc = False
            type_string = ''
            for line in f_string_array:
                line_long_enough = False
                thick_in_line = False
                if len(line) > 1:
                    line_long_enough = True
                if "TRANSMISSION" in line:
                    captype = False
                    type_string = 'trans'
                if "CAPTURE DATA" in line:
                    captype = True
                    type_string = 'capt'
                if "(NUC./B)" in line:
                    start_of_mc = True
                    continue
                if "M O N T E" in line:
                    start_of_mc = False
                if line_long_enough and start_of_mc and ('.' in line[1:10]):
                    thick_in_line = True
                    thickness = line[0:10]
                if line_long_enough and start_of_mc and ('I N P U T' in line):
                    # break if the end of the MC region is found
                    break
                
                if line_long_enough and start_of_mc:
                    # energy line -- begin new_line to write 
                    if "." == line[13]:
                        # initialize string variables
                        E_n,av_cap_xs,av_tot_xs,ss_cor_fc,t = '','','','',''
                        new_line = ''
                        E_n = line[6:19]
                        # sample thickness takes up extra columns, so read this one differently
                        if thick_in_line:
                            E_n = '    '+line[10:19]
                        av_cap_xs,av_tot_xs = line[19:28],line[29:38]
                        if captype:
                            ss_cor_fc = line[65:74]
                            t = line[119:128]
                        else:
                            ss_cor_fc = line[51:61]
                            t = line[39:48]

                    # collect the uncertainty information from next line and place in new_line
                    if "." != line[13]:
                        # initialize string variables
                        dav_cap_xs,dav_tot_xs,dss_cor_fc,dt = '','','',''
                        dav_cap_xs,dav_tot_xs = line[19:28],line[29:38]
                        if captype:
                            dss_cor_fc = line[65:74]
                            dt = line[119:128]
                        else:
                            dss_cor_fc = line[51:61]
                            dt = line[39:48]
                    
                    
                    if "." != line[13]:
                        new_line += E_n+' '+av_tot_xs+' '+dav_tot_xs+' '+av_cap_xs+' '+dav_cap_xs
                        new_line += ' '+ss_cor_fc+' '+dss_cor_fc
                        new_line += ' '+t+' '+dt+' '+thickness+' '+type_string
                    # write the new_line into the new clean file when the 1st and 2nd line
                    # for the energy has been read
                    if "." != line[13]:
                        fnew.write(new_line+"\n")


